# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import orjson


def _load_views():
    try:
        with open("views.json", "rb") as defs:
            return orjson.loads(defs.read())
    except Exception as err:  # nosec
        if not err:
            pass
        # DEBUG:: log (f"[OPTERYX] Unable to open views definition file. {err}")
        return {}


VIEWS = _load_views()


def is_view(view_name: str) -> bool:
    return view_name in VIEWS


def view_as_plan(view_name: str):
    from opteryx.planner.logical_planner import do_logical_planning_phase
    from opteryx.third_party import sqloxide
    from opteryx.utils.sql import clean_statement
    from opteryx.utils.sql import remove_comments

    operation = VIEWS.get(view_name)["statement"]

    clean_sql = clean_statement(remove_comments(operation))
    parsed_statements = sqloxide.parse_sql(clean_sql, dialect="mysql")
    logical_plan, _, _ = next(do_logical_planning_phase(parsed_statements))

    return logical_plan



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List
from typing import Optional

import numpy
import pyarrow
from pyarrow import compute

from opteryx.exceptions import SqlError


def list_contains(array, item):
    """
    does array contain item
    """
    if array is None:
        return False
    return item in set(array)


def list_contains_any(array: numpy.ndarray, items: numpy.ndarray) -> numpy.ndarray:
    """
    Check if any element in each sub-array of 'array' is present in 'items'.

    Parameters:
        array (numpy.ndarray): A 1D array where each element is an iterable (e.g., list, set).
        items (numpy.ndarray): An array of items to check against each sub-array in 'array'.

    Returns:
        numpy.ndarray: A boolean array where each element indicates whether any item
                       from 'items' is found in the corresponding sub-array of 'array'.
    """
    items_set = set(items[0])
    res = numpy.empty(array.size, dtype=bool)
    for i, test_set in enumerate(array):
        # Using not to correctly capture overlap (isdisjoint is True when no common elements)
        res[i] = not items_set.isdisjoint(test_set) if test_set is not None else False
    return res


def list_contains_all(array, items):
    """
    does array contain all of the items in items
    """
    if array is None:
        return False
    return set(array).issuperset(items)


def search(array, item, ignore_case: Optional[List[bool]] = None):
    """
    `search` provides a way to look for values across different field types, rather
    than doing a LIKE on a string, IN on a list, `search` adapts to the field type.

    This performs a pre-filter of the data to remove nulls - this means that the
    checks should generally be faster.
    """

    if ignore_case is None:
        ignore_case = [True]

    item = item[0]  # [#325]
    record_count = array.size

    if record_count > 0:
        null_positions = compute.is_null(array, nan_is_null=True)
        # if all the values are null, short-cut
        if null_positions.false_count == 0:
            return numpy.full(record_count, False, numpy.bool_)
        # do we have any nulls?
        compressed = null_positions.true_count > 0
        null_positions = numpy.invert(null_positions)
        # remove nulls from the checks
        if compressed:
            array = array.compress(null_positions)
        array_type = type(array[0])
    else:
        return numpy.array([False], dtype=numpy.bool_)

    if array_type in (str, bytes):
        # return True if the value is in the string
        results_mask = compute.match_substring(array, pattern=item, ignore_case=ignore_case[0])
    elif array_type == numpy.ndarray:
        # converting to a set is faster for a handful of items which is what we're
        # almost definitely working with here - note compute.index is about 50x slower
        results_mask = numpy.array([item in set(record) for record in array], dtype=numpy.bool_)
    elif array_type == dict:
        results_mask = numpy.array([item in record.values() for record in array], dtype=numpy.bool_)
    else:
        raise SqlError("SEARCH can only be used with VARCHAR, BLOB, LIST and STRUCT.")

    if compressed:
        # fill the result set
        results = numpy.full(record_count, False, numpy.bool_)
        results[numpy.nonzero(null_positions)] = results_mask
        return results

    return results_mask


def iif(mask, true_values, false_values):
    # we have three columns, the first is TRUE offsets
    # the second is TRUE response
    # the third is FAST response

    if isinstance(mask, pyarrow.lib.BooleanArray) or (
        isinstance(mask, numpy.ndarray) and mask.dtype == numpy.bool_
    ):
        mask = numpy.nonzero(mask)[0]

    response = false_values

    for index in mask:
        response[index] = true_values[index]
    return response


def if_null(values, replacement):
    """
    Replace null values in the input array with corresponding values from the replacement array.

    Parameters:
        values: Union[numpy.ndarray, pyarrow.Array]
            The input array which may contain null values.
        replacement: numpy.ndarray
            The array with replacement values corresponding to the null positions in the input array.

    Returns:
        numpy.ndarray
            The input array with null values replaced by corresponding values from the replacement array.
    """
    from opteryx.managers.expression.unary_operations import _is_null

    # Check if the values array is a pyarrow array and convert it to a numpy array if necessary
    if isinstance(values, pyarrow.Array):
        values = values.to_numpy(False)
    if isinstance(values, list):
        values = numpy.array(values)

    response = values.copy()  # Create a copy of the array to avoid modifying the original
    is_null_array = _is_null(values)
    for index, is_null in enumerate(is_null_array):
        if is_null:
            response[index] = replacement[index]

    return response


def null_if(col1, col2):
    return [None if a == b else a for a, b in zip(col1, col2)]


def case_when(conditions, values):
    res = []
    cons = list(zip(*conditions))
    vals = zip(*values)
    for idx, val_set in enumerate(vals):
        offset = next((i for i, j in enumerate(cons[idx]) if j), None)
        if offset is not None:
            res.append(val_set[offset])
        else:
            res.append(None)
    return res


def cosine_similarity(arr, val):
    """
    ad hoc cosine similarity function, slow.
    """

    from opteryx.compiled.functions import tokenize_and_remove_punctuation
    from opteryx.compiled.functions import vectorize
    from opteryx.virtual_datasets.stop_words import STOP_WORDS

    def cosine_similarity(
        vec1: numpy.ndarray, vec2: numpy.ndarray, vec2_norm: numpy.float32
    ) -> float:
        vec1 = vec1.astype(numpy.float32)
        vec1_norm = numpy.linalg.norm(vec1)

        product = vec1_norm * vec2_norm
        if product == 0:
            return 0

        return numpy.dot(vec1, vec2) / product

    # import time

    if len(val) == 0:
        return []
    tokenized_literal = tokenize_and_remove_punctuation(str(val[0]), STOP_WORDS)
    if len(tokenized_literal) == 0:
        return [0.0] * len(arr)
    # print(len(val))

    # t = time.monotonic_ns()
    tokenized_strings = [tokenize_and_remove_punctuation(s, STOP_WORDS) for s in arr] + [
        tokenized_literal
    ]
    # print("time tokenizing ", time.monotonic_ns() - t)
    # t = time.monotonic_ns()
    vectors = [vectorize(tokens) for tokens in tokenized_strings]
    # print("time vectorizing", time.monotonic_ns() - t)
    comparison_vector = vectors[-1].astype(numpy.float32)
    comparison_vector_norm = numpy.linalg.norm(comparison_vector)

    if comparison_vector_norm == 0.0:
        return [0.0] * len(val)

    # t = time.monotonic_ns()
    similarities = [
        cosine_similarity(vector, comparison_vector, comparison_vector_norm)
        for vector in vectors[:-1]
    ]
    # print("time comparing  ", time.monotonic_ns() - t)

    return similarities


def jsonb_object_keys(arr):
    if len(arr) == 0:
        return []
    if isinstance(arr[0], dict):
        result = [[str(key) for key in row] for row in arr]
    if isinstance(arr[0], (str, bytes)):
        import simdjson

        def keys(doc):
            return simdjson.Parser().parse(doc).keys()  # type:ignore

        result = [[str(key) for key in keys(row)] for row in arr]
    return pyarrow.array(result)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Function Catalogue
"""

import inspect
import re
import typing
from dataclasses import dataclass
from enum import Enum
from enum import auto
from functools import wraps
from typing import Any
from typing import Callable
from typing import Dict
from typing import Generic
from typing import List
from typing import Tuple
from typing import TypeVar
from typing import Union

from orso.types import PYTHON_TO_ORSO_MAP
from orso.types import OrsoTypes

PYTHON_TO_ORSO_MAP[typing.Any] = "INPUT TYPE"


def not_implemented(*args, **kwds):
    raise NotImplementedError("Subclasses must implement the _func method.")


class ParameterMode(Enum):
    SCALAR = auto()
    ARRAY = auto()


T = TypeVar("T")  # Declare type variable


@dataclass
class Parameter(Generic[T]):
    def __init__(
        self,
        types: List[OrsoTypes],
        default: Any,
        description: str = None,
        mode: ParameterMode = ParameterMode.ARRAY,
        validator: str = r"^.*$",
    ):
        self.default = default
        self.types = types
        self.description = description
        self.mode = mode
        self.validator = re.compile(validator) if validator else None

    def __repr__(self):
        return f"<Parameter (type={[a.name for a in self.types]}, default={self.default})>"


class FunctionMode(Enum):
    # Aggregation functions accept a column of values and return a single value
    AGGREGATION = auto()
    FUNCTION = auto()
    # Constant functions return a single value, usually with no input
    CONSTANT = auto()
    ALIAS = auto()


class Function:
    func = not_implemented
    mode: FunctionMode = FunctionMode.FUNCTION
    cost: Union[float]
    return_type: OrsoTypes
    attributes: Dict[str, OrsoTypes]


def determine_return_types(function):
    """
    Use introspection to work out what the return type of a function
    will be.

    We only support discrete types
    """
    return_type_hints = typing.get_type_hints(function).get("return")
    if return_type_hints is not None:
        return PYTHON_TO_ORSO_MAP.get(return_type_hints, "OTHER")
    return "UNKNOWN"


def determine_argument_types(function):
    func_signature = inspect.signature(function)
    func_parameters = func_signature.parameters
    type_hints = typing.get_type_hints(function)

    parameters = {}
    for arg_name, arg_value in func_parameters.items():
        arg_type = type_hints.get(arg_name)
        value = arg_value.default
        if not isinstance(value, Parameter):
            value = Parameter(default=value, types=[PYTHON_TO_ORSO_MAP.get(arg_type, "OTHER")])

        parameters[arg_name] = value
    return parameters


class _FunctionCatalogue:
    def __init__(self):
        self.function_catalogue: List[Tuple[str, dict]] = []

    def get(self, func: str, parameters=None):
        """return a function with a given name"""
        func = func.upper()
        candidates = [(name, spec) for name, spec in self.function_catalogue if name == func]
        if len(candidates) == 0:
            return None
        # do type checks
        return candidates[0][1]["function"]

    def full_details(self, func: str):
        pass
        # include the parameter types and names and the return type

    def suggest(self, func):
        """return the function with the nearest name match"""
        from itertools import permutations

        from opteryx.utils import suggest_alternative

        available_functions = set(name for name, spec in self.function_catalogue)

        # try a fuzzy search (typos) first
        suggestion = suggest_alternative(func, available_functions)

        # if it probably wasn't a typo, try rearranging the parts of the function names
        if suggestion is None:
            parts = func.split("_")
            combinations = permutations(parts)
            for combination in combinations:
                candidate = "_".join(combination)
                suggestion = suggest_alternative(candidate, available_functions)
                if suggestion:
                    break

        return suggestion

    def collect(self):
        function_list = []
        for function, specification in self.function_catalogue:
            function_list.append({"name": function, **specification})
        return function_list

    def function_collection(self):
        import pyarrow

        collection = [
            {
                k: str(v)
                for k, v in f.items()
                if k in {"name", "description", "mode", "return_type", "parameters"}
            }
            for f in self.collect()
        ]

        return pyarrow.Table.from_pylist(collection)

    def __call__(self, mode: FunctionMode, **metadata) -> Callable:
        def decorator(func: Callable) -> Callable:
            @wraps(func)
            def wrapper(*args, **kwargs):
                # Placeholder for parameter specs; to be replaced with actual parameter determination
                parameters_spec = determine_argument_types(func)

                new_args = []
                new_kwargs = {}

                # Adjust positional arguments based on the determined parameters_spec
                for i, arg in enumerate(args):
                    if i < len(parameters_spec):
                        param_name = list(parameters_spec.keys())[i]
                        param_spec = parameters_spec[param_name]
                        if param_spec.mode == ParameterMode.SCALAR and isinstance(arg, list):
                            arg = arg[0]  # Simplified example, assumes non-empty lists
                        elif param_spec.mode == ParameterMode.ARRAY and not isinstance(arg, list):
                            arg = [arg]
                    new_args.append(arg)

                # Similar adjustment for keyword arguments
                for key, value in kwargs.items():
                    param_spec = parameters_spec.get(key)
                    if (
                        param_spec
                        and param_spec.mode == ParameterMode.SCALAR
                        and isinstance(value, list)
                    ):
                        value = value[0]  # Simplified example
                    elif (
                        param_spec
                        and param_spec.mode == ParameterMode.ARRAY
                        and not isinstance(value, list)
                    ):
                        value = [value]
                    new_kwargs[key] = value

                return func(*new_args, **new_kwargs)

            # Register the original function with the wrapped one
            returns = determine_return_types(func)
            parameters = determine_argument_types(func)
            self.function_catalogue.append(
                (
                    func.__name__.upper(),
                    {
                        "function": wrapper,  # Store the wrapper instead of the original function
                        "mode": mode,
                        "description": metadata.get("description", func.__doc__),
                        "return_type": returns,
                        "parameters": parameters,
                        **metadata,
                    },
                )
            )
            return wrapper  # Return the wrapped function for use

        return decorator


function_catalogue = _FunctionCatalogue()

if __name__ == "__main__":

    @function_catalogue(mode=FunctionMode.CONSTANT)
    def pi() -> float:
        """Irational constant Pi"""
        return 3.14

    # Example usage
    @function_catalogue(mode=FunctionMode.AGGREGATION)
    def example_function(
        x: Parameter[int] = Parameter(
            default=12, types=[OrsoTypes.INTEGER], mode=ParameterMode.SCALAR
        ),
        y: Parameter[int] = Parameter(
            default=None, types=[OrsoTypes.INTEGER], mode=ParameterMode.SCALAR
        ),
    ) -> int:
        """Example function that adds two numbers."""
        return x + y  # type: ignore

    print(function_catalogue.collect())
    print(function_catalogue.function_collection())
    fun = function_catalogue.get("example_function")
    print(function_catalogue.suggest("function_example"))
    print(function_catalogue.suggest("function_examp"))

    print(fun(1, 2))



from .boolean_simplication import BooleanSimplificationStrategy
from .constant_folding import ConstantFoldingStrategy
from .distinct_pushdown import DistinctPushdownStrategy
from .operator_fusion import OperatorFusionStrategy
from .predicate_pushdown import PredicatePushdownStrategy
from .predicate_rewriter import PredicateRewriteStrategy
from .projection_pushdown import ProjectionPushdownStrategy
from .redundant_operators import RedundantOperationsStrategy
from .split_conjunctive_predicates import SplitConjunctivePredicatesStrategy

__all__ = [
    "BooleanSimplificationStrategy",
    "ConstantFoldingStrategy",
    "DistinctPushdownStrategy",
    "OperatorFusionStrategy",
    "PredicatePushdownStrategy",
    "PredicateRewriteStrategy",
    "ProjectionPushdownStrategy",
    "RedundantOperationsStrategy",
    "SplitConjunctivePredicatesStrategy",
]



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
from typing import Callable
from typing import List
from typing import Optional

from opteryx.managers.schemes import BasePartitionScheme


class DefaultPartitionScheme(BasePartitionScheme):
    def get_blobs_in_partition(
        self,
        *,
        blob_list_getter: Callable,
        prefix: str,
        start_date: Optional[datetime.datetime],
        end_date: Optional[datetime.datetime],
        **kwargs,
    ) -> List[str]:
        return blob_list_getter(prefix=prefix)



# cython: language_level=3

from libc.stdlib cimport malloc, free
import numpy as np
cimport numpy as cnp
cimport cython
from libc.stdint cimport int32_t

@cython.boundscheck(False)
@cython.wraparound(False)
cpdef tuple build_rows_indices_and_column(cnp.ndarray column_data):
    cdef int32_t i, total_size = 0
    cdef int32_t length
    cdef list flat_data
    cdef int32_t row_count = len(column_data)
    cdef int32_t *lengths = <int32_t *>malloc(row_count * sizeof(int32_t))
    if lengths is NULL:
        raise MemoryError("Failed to allocate memory for lengths array.")

    for i in range(row_count):
        length = len(column_data[i])
        lengths[i] = length
        total_size += length

    if total_size == 0:
        free(lengths)
        return (np.array([], dtype=np.int32), np.array([], dtype=object))

    flat_data = [''] * total_size
    cdef int32_t *indices = <int32_t *>malloc(total_size * sizeof(int32_t))
    if indices is NULL:
        raise MemoryError("Failed to allocate memory.")

    cdef int32_t start = 0
    cdef int32_t end = 0
    cdef int32_t j = 0

    for i in range(row_count):
        end = start + lengths[i]
        for j in range(start, end):
            indices[j] = i 
            flat_data[j] = column_data[i][j - start]
        start = end
    free(lengths)

    cdef cnp.int32_t[:] mv = <cnp.int32_t[:total_size]>indices
    # Create a NumPy array that is a copy of the memoryview, 
    # which in turn makes it safe to free the original indices memory.
    np_array = np.array(mv, copy=True)
    free(indices)  # Now it's safe to free indices since np_array has its own copy.
    return (np_array, flat_data)


from libc.stdlib cimport malloc, realloc, free
import numpy as np
cimport numpy as cnp
cimport cython
from libc.stdint cimport int32_t

@cython.boundscheck(False)
@cython.wraparound(False)
cpdef tuple build_filtered_rows_indices_and_column(cnp.ndarray column_data, set valid_values):
    """
    Build row indices and flattened column data for matching values from a column of array-like elements.

    Parameters:
        column_data: ndarray
            An array of arrays from which to create row indices and flattened data.
        valid_values: set
            A set of values to filter the rows by during the cross join.

    Returns:
        tuple of (ndarray, list)
            Returns a tuple containing an array of indices and a list of flattened data for rows that match the filter.
    """
    cdef int32_t i, index = 0, allocated_size
    cdef int32_t row_count = len(column_data)
    cdef int32_t initial_alloc_size = 500
    allocated_size = initial_alloc_size
    cdef int32_t *indices = <int32_t *>malloc(allocated_size * sizeof(int32_t))
    cdef list flat_data = [None] * allocated_size
    if indices is NULL:
        raise MemoryError("Failed to allocate memory for indices.")

    for i in range(row_count):
        for value in column_data[i]:
            if value in valid_values:
                if index == allocated_size:  # Check if we need to expand the memory allocation
                    allocated_size += initial_alloc_size
                    indices = <int32_t *>realloc(indices, allocated_size * sizeof(int32_t))
                    flat_data.extend([None] * initial_alloc_size)  # Extend flat_data by the same amount
                    if indices is NULL:
                        raise MemoryError("Failed to reallocate memory for indices.")
                flat_data[index] = value
                indices[index] = i
                index += 1

    if index == 0:
        free(indices)
        return (np.array([], dtype=np.int32), [])

    # Slice to actual used size before creating final arrays
    cdef int32_t[:] indices_view = <int32_t[:index]>indices
    cdef cnp.ndarray final_indices = np.array(indices_view, dtype=np.int32)
    free(indices)  # Free the original buffer now

    return (final_indices, flat_data[:index])



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Outer Join Node

This is a SQL Query Execution Plan Node.

PyArrow has LEFT/RIGHT/FULL OUTER JOIN implementations, but they error when the
relations being joined contain STRUCT or ARRAY columns so we've written our own
OUTER JOIN implementations.

We also have our own INNER JOIN implementations, it's really just the less
popular SEMI and ANTI joins we leave to PyArrow for now.
"""

import time
from typing import Generator
from typing import List

import pyarrow

from opteryx.compiled.structures import HashTable
from opteryx.models import QueryProperties
from opteryx.operators import BasePlanNode
from opteryx.operators import OperatorType
from opteryx.utils.arrow import align_tables


def left_join(left_relation, right_relation, left_columns: List[str], right_columns: List[str]):
    """
    Perform an LEFT JOIN.

    This implementation ensures that all rows from the left table are included in the result set,
    with rows from the right table matched where possible, and columns from the right table
    filled with NULLs where no match is found.

    Parameters:
        left_relation (pyarrow.Table): The left pyarrow.Table to join.
        right_relation (pyarrow.Table): The right pyarrow.Table to join.
        left_columns (list of str): Column names from the left table to join on.
        right_columns (list of str): Column names from the right table to join on.

    Returns:
        A pyarrow.Table containing the result of the LEFT JOIN operation.
    """
    from collections import deque

    from opteryx.compiled.structures.hash_table import hash_join_map

    left_indexes = deque()
    right_indexes = deque()

    right_relation = pyarrow.concat_tables(right_relation.execute(), promote_options="none")
    right_hash = hash_join_map(right_relation, right_columns)

    for left_batch in left_relation.execute():
        left_hash = hash_join_map(left_batch, left_columns)
        for hash_value, left_rows in left_hash.hash_table.items():
            right_rows = right_hash.get(hash_value)
            if right_rows:
                for l in left_rows:
                    for r in right_rows:
                        left_indexes.append(l)
                        right_indexes.append(r)
            else:
                for l in left_rows:
                    left_indexes.append(l)
                    right_indexes.append(None)

            if len(left_indexes) > 50_000:
                table = align_tables(
                    right_relation, left_batch, list(right_indexes), list(left_indexes)
                )
                yield table
                left_indexes.clear()
                right_indexes.clear()

        if len(left_indexes) > 0:
            table = align_tables(
                right_relation, left_batch, list(right_indexes), list(left_indexes)
            )
            yield table
            left_indexes.clear()
            right_indexes.clear()


def full_join(left_relation, right_relation, left_columns: List[str], right_columns: List[str]):
    chunk_size = 1000
    right_relation = pyarrow.concat_tables(right_relation.execute(), promote_options="none")

    hash_table = HashTable()
    non_null_right_values = right_relation.select(right_columns).itercolumns()
    for i, value_tuple in enumerate(zip(*non_null_right_values)):
        hash_table.insert(hash(value_tuple), i)

    left_indexes = []
    right_indexes = []

    left_relation = pyarrow.concat_tables(left_relation.execute(), promote_options="none")
    left_values = left_relation.select(left_columns).itercolumns()
    for i, value_tuple in enumerate(zip(*left_values)):
        rows = hash_table.get(hash(value_tuple))
        if rows:
            right_indexes.extend(rows)
            left_indexes.extend([i] * len(rows))
        else:
            right_indexes.append(None)
            left_indexes.append(i)

    for i in range(right_relation.num_rows):
        if i not in right_indexes:
            right_indexes.append(i)
            left_indexes.append(None)

    for i in range(0, len(left_indexes), chunk_size):
        chunk_left_indexes = left_indexes[i : i + chunk_size]
        chunk_right_indexes = right_indexes[i : i + chunk_size]

        # Align this chunk and add the resulting table to our list
        yield align_tables(right_relation, left_relation, chunk_right_indexes, chunk_left_indexes)


def right_join(left_relation, right_relation, left_columns: List[str], right_columns: List[str]):
    """
    Perform a RIGHT JOIN.

    This implementation ensures that all rows from the right table are included in the result set,
    with rows from the left table matched where possible, and columns from the left table
    filled with NULLs where no match is found.

    Parameters:
        left_relation (pyarrow.Table): The left pyarrow.Table to join.
        right_relation (pyarrow.Table): The right pyarrow.Table to join.
        left_columns (list of str): Column names from the left table to join on.
        right_columns (list of str): Column names from the right table to join on.

    Yields:
        pyarrow.Table: A chunk of the result of the RIGHT JOIN operation.
    """
    chunk_size = 1000
    left_relation = pyarrow.concat_tables(left_relation.execute(), promote_options="none")

    hash_table = HashTable()
    non_null_left_values = left_relation.select(left_columns).itercolumns()
    for i, value_tuple in enumerate(zip(*non_null_left_values)):
        hash_table.insert(hash(value_tuple), i)

    # Iterate over the right_relation in chunks
    right_batches = right_relation.execute()
    for right_batch in right_batches:
        for right_chunk in right_batch.to_batches(chunk_size):
            left_indexes = []
            right_indexes = []

            right_values = right_chunk.select(right_columns).itercolumns()
            for i, value_tuple in enumerate(zip(*right_values)):
                rows = hash_table.get(hash(value_tuple))
                if rows:
                    left_indexes.extend(rows)
                    right_indexes.extend([i] * len(rows))
                else:
                    left_indexes.append(None)
                    right_indexes.append(i)

            # Yield the aligned chunk
            # we intentionally swap them to the other calls so we're building a table
            # not a record batch (what the chunk is)
            yield align_tables(left_relation, right_chunk, left_indexes, right_indexes)


def left_anti_join(
    left_relation, right_relation, left_columns: List[str], right_columns: List[str]
):
    """
    Perform a LEFT ANTI JOIN.

    This implementation ensures that all rows from the left table are included in the result set,
    where there are no matching rows in the right table based on the join columns.

    Parameters:
        left_relation (pyarrow.Table): The left pyarrow.Table to join.
        right_relation (pyarrow.Table): The right pyarrow.Table to join.
        left_columns (list of str): Column names from the left table to join on.
        right_columns (list of str): Column names from the right table to join on.

    Returns:
        A pyarrow.Table containing the result of the LEFT ANTI JOIN operation.
    """
    right_relation = pyarrow.concat_tables(right_relation.execute(), promote_options="none")

    hash_table = HashTable()
    non_null_right_values = right_relation.select(right_columns).itercolumns()
    for i, value_tuple in enumerate(zip(*non_null_right_values)):
        hash_table.insert(hash(value_tuple), i)

    at_least_once = False
    # Iterate over the left_relation in chunks
    for left_batch in left_relation.execute():
        left_indexes = []
        left_values = left_batch.select(left_columns).itercolumns()
        for i, value_tuple in enumerate(zip(*left_values)):
            rows = hash_table.get(hash(value_tuple))
            if not rows:  # Only include left rows that have no match in the right table
                left_indexes.append(i)

        # Filter the left_chunk based on the anti join condition
        if left_indexes:
            yield left_batch.take(left_indexes)
            at_least_once = True

    if not at_least_once:
        yield left_batch.slice(0, 0)


def left_semi_join(
    left_relation, right_relation, left_columns: List[str], right_columns: List[str]
):
    """
    Perform a LEFT SEMI JOIN.

    This implementation ensures that all rows from the left table that have a matching row in the right table
    based on the join columns are included in the result set.

    Parameters:
        left_relation (pyarrow.Table): The left pyarrow.Table to join.
        right_relation (pyarrow.Table): The right pyarrow.Table to join.
        left_columns (list of str): Column names from the left table to join on.
        right_columns (list of str): Column names from the right table to join on.

    Returns:
        A pyarrow.Table containing the result of the LEFT SEMI JOIN operation.
    """
    right_relation = pyarrow.concat_tables(right_relation.execute(), promote_options="none")

    hash_table = HashTable()
    non_null_right_values = right_relation.select(right_columns).itercolumns()
    for i, value_tuple in enumerate(zip(*non_null_right_values)):
        hash_table.insert(hash(value_tuple), i)

    at_least_once = False
    # Iterate over the left_relation in chunks
    for left_batch in left_relation.execute():
        left_indexes = []
        left_values = left_batch.select(left_columns).itercolumns()

        for i, value_tuple in enumerate(zip(*left_values)):
            rows = hash_table.get(hash(value_tuple))
            if rows:  # Only include left rows that have a match in the right table
                left_indexes.append(i)

        # Filter the left_chunk based on the anti join condition
        if left_indexes:
            yield left_batch.take(left_indexes)
            at_least_once = True

    if not at_least_once:
        yield left_batch.slice(0, 0)


class OuterJoinNode(BasePlanNode):
    operator_type = OperatorType.PASSTHRU

    def __init__(self, properties: QueryProperties, **config):
        super().__init__(properties=properties)
        self._join_type = config["type"]
        self._on = config.get("on")
        self._using = config.get("using")

        self._left_columns = config.get("left_columns")
        self._left_relation = config.get("left_relation_names")

        self._right_columns = config.get("right_columns")
        self._right_relation = config.get("right_relation_names")

    @classmethod
    def from_json(cls, json_obj: str) -> "BasePlanNode":  # pragma: no cover
        raise NotImplementedError()

    @property
    def name(self):  # pragma: no cover
        return self._join_type

    @property
    def config(self):  # pragma: no cover
        from opteryx.managers.expression import format_expression

        if self._on:
            return f"{self._join_type.upper()} JOIN ({format_expression(self._on, True)})"
        if self._using:
            return f"{self._join_type.upper()} JOIN (USING {','.join(map(format_expression, self._using))})"
        return f"{self._join_type.upper()}"

    def execute(self) -> Generator:
        left_node = self._producers[0]  # type:ignore
        right_node = self._producers[1]  # type:ignore

        join_provider = providers.get(self._join_type)

        start = time.monotonic_ns()
        for morsel in join_provider(
            left_relation=left_node,
            right_relation=right_node,
            left_columns=self._left_columns,
            right_columns=self._right_columns,
        ):
            self.statistics.time_outer_join += time.monotonic_ns() - start
            yield morsel
            start = time.monotonic_ns()


providers = {
    "left outer": left_join,
    "full outer": full_join,
    "right outer": right_join,
    "left anti": left_anti_join,
    "left semi": left_semi_join,
}



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Join Node

We have our own implementations of INNER and OUTER joins, this uses PyArrow
to implement less-common joins of ANTI and SEMI joins.
"""

from typing import Generator

import pyarrow

from opteryx.exceptions import UnsupportedSyntaxError
from opteryx.models import QueryProperties
from opteryx.operators import BasePlanNode
from opteryx.operators import OperatorType


class JoinNode(BasePlanNode):
    operator_type = OperatorType.PASSTHRU

    def __init__(self, properties: QueryProperties, **config):
        super().__init__(properties=properties)
        self._join_type = config["type"]
        self._on = config.get("on")
        self._using = config.get("using")

        self._left_columns = config.get("left_columns")
        self._left_relation = config.get("left_relation_names")

        self._right_columns = config.get("right_columns")
        self._right_relation = config.get("right_relation_names")

    @classmethod
    def from_json(cls, json_obj: str) -> "BasePlanNode":  # pragma: no cover
        raise NotImplementedError()

    @property
    def name(self):  # pragma: no cover
        return f"{self._join_type} Join"

    @property
    def config(self):  # pragma: no cover
        from opteryx.managers.expression import format_expression

        if self._on:
            return f"{self._join_type.upper()} JOIN ({format_expression(self._on, True)})"
        if self._using:
            return f"{self._join_type.upper()} JOIN (USING {','.join(map(format_expression, self._using))})"
        return f"{self._join_type.upper()}"

    def execute(self) -> Generator:
        left_node = self._producers[0]  # type:ignore
        right_node = self._producers[1]  # type:ignore

        left_table = pyarrow.concat_tables(left_node.execute(), promote_options="none")
        right_table = pyarrow.concat_tables(right_node.execute(), promote_options="none")

        try:
            new_morsel = left_table.join(
                right_table,
                keys=self._left_columns,
                right_keys=self._right_columns,
                join_type=self._join_type,
                coalesce_keys=self._using is not None,
            )
        except pyarrow.ArrowInvalid as err:  # pragma: no cover
            last_token = str(err).split(" ")[-1]
            column = None
            for col in left_node.columns:
                if last_token == col.identity:
                    column = col.name
                    break
            for col in right_node.columns:
                if last_token == col.identity:
                    column = col.name
                    break
            if column:
                raise UnsupportedSyntaxError(
                    f"Unable to ANTI/SEMI JOIN with unsupported column types in table, '{column}'."
                ) from err
            raise UnsupportedSyntaxError(
                "Unable to ANTI/SEMI JOIN with unsupported column types in table."
            ) from err

        yield new_morsel



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This implements an interface to REDIS

If we have 10 failures in a row, stop trying to use the cache.
"""

import os
from typing import Union

from orso.tools import single_item_cache

from opteryx.exceptions import MissingDependencyError
from opteryx.managers.kvstores import BaseKeyValueStore

MAXIMUM_CONSECUTIVE_FAILURES: int = 10


@single_item_cache
def _redis_server(**kwargs):
    """
    Handling connecting to REDIS
    """
    # the server must be set in the environment
    redis_config = kwargs.get("server", os.environ.get("REDIS_CONNECTION"))
    if redis_config is None:
        return None

    try:
        import redis
    except ImportError as err:
        raise MissingDependencyError(err.name) from err

    return redis.from_url(redis_config)


class RedisCache(BaseKeyValueStore):
    """
    Cache object
    """

    def __init__(self, **kwargs):
        """
        Parameters:
            server: string (optional)
                Sets the memcached server and port (server:port). If not provided
                the value will be obtained from the OS environment.
        """
        self._server = _redis_server(**kwargs)
        if self._server is None:
            self._consecutive_failures: int = MAXIMUM_CONSECUTIVE_FAILURES
        else:
            self._consecutive_failures = 0
        self.hits: int = 0
        self.misses: int = 0
        self.skips: int = 0
        self.errors: int = 0
        self.sets: int = 0

    def get(self, key: bytes) -> Union[bytes, None]:
        if self._consecutive_failures >= MAXIMUM_CONSECUTIVE_FAILURES:
            self.skips += 1
            return None
        try:
            response = self._server.get(key)
            self._consecutive_failures = 0
            if response:
                self.hits += 1
                return bytes(response)
        except Exception as err:
            self._consecutive_failures += 1
            if self._consecutive_failures >= MAXIMUM_CONSECUTIVE_FAILURES:
                import datetime

                print(
                    f"{datetime.datetime.now()} [CACHE] Disabling remote Redis cache due to persistent errors ({err})."
                )
            self.errors += 1
            return None

        self.misses += 1
        return None

    def set(self, key: bytes, value: bytes) -> None:
        if self._consecutive_failures < MAXIMUM_CONSECUTIVE_FAILURES:
            try:
                self._server.set(key, value)
                self.sets += 1
            except Exception as err:
                # if we fail to set, stop trying
                self._consecutive_failures = MAXIMUM_CONSECUTIVE_FAILURES
                self.errors += 1
                import datetime

                print(
                    f"{datetime.datetime.now()} [CACHE] Disabling remote Redis cache due to persistent errors ({err}) [SET]."
                )
        else:
            self.skips += 1

    def __del__(self):
        pass
        # DEBUG: log(f"Redis <hits={self.hits} misses={self.misses} sets={self.sets} skips={self.skips} errors={self.errors}>")



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Optimization Rule - Constant Folding

Type: Heuristic
Goal: Evaluate Once

We identify branches in expressions where there are no identifiers, these usually
mean we can evaluate them once, in the optimization phase, and replace them with a
constant for handling in the execution phase, reducing the amount of work done by
the execution engine.

We run this strategy twice, once at the beginning, which primarily handles user
entered expressions we can optimize, and again at the end which handles where
we've rewritten expressions at part of other optimizations which can be folded.
"""

import datetime
from typing import Any

import numpy
from orso.types import OrsoTypes

from opteryx.managers.expression import NodeType
from opteryx.managers.expression import evaluate
from opteryx.managers.expression import get_all_nodes_of_type
from opteryx.models import Node
from opteryx.planner.logical_planner import LogicalPlan
from opteryx.planner.logical_planner import LogicalPlanNode
from opteryx.planner.logical_planner import LogicalPlanStepType
from opteryx.virtual_datasets import no_table_data

from .optimization_strategy import OptimizationStrategy
from .optimization_strategy import OptimizerContext


def build_literal_node(value: Any, root: Node, suggested_type: OrsoTypes):
    """
    Build a literal node with the appropriate type based on the value.
    """
    # Convert value if it has `as_py` method (e.g., from PyArrow)
    if hasattr(value, "as_py"):
        value = value.as_py()

    if value is None:
        # Matching None has complications
        root.value = None
        root.node_type = NodeType.LITERAL
        root.type = OrsoTypes.NULL
        root.left = None
        root.right = None
        return root

    # Define a mapping of types to OrsoTypes
    type_mapping = {
        bool: OrsoTypes.BOOLEAN,
        numpy.bool_: OrsoTypes.BOOLEAN,
        str: OrsoTypes.VARCHAR,
        numpy.str_: OrsoTypes.VARCHAR,
        bytes: OrsoTypes.BLOB,
        numpy.bytes_: OrsoTypes.BLOB,
        int: OrsoTypes.INTEGER,
        numpy.int64: OrsoTypes.INTEGER,
        float: OrsoTypes.DOUBLE,
        numpy.float64: OrsoTypes.DOUBLE,
        numpy.datetime64: OrsoTypes.TIMESTAMP,
        datetime.datetime: OrsoTypes.TIMESTAMP,
        datetime.time: OrsoTypes.TIME,
        datetime.date: OrsoTypes.DATE,
    }

    value_type = type(value)
    # Determine the type from the value using the mapping
    if value_type in type_mapping or suggested_type not in (OrsoTypes._MISSING_TYPE, 0, None):
        root.value = value
        root.node_type = NodeType.LITERAL
        root.type = (
            suggested_type
            if suggested_type not in (OrsoTypes._MISSING_TYPE, 0, None)
            else type_mapping[value_type]
        )
        root.left = None
        root.right = None

    # DEBUG:log (f"Unable to create literal node for {value}, of type {value_type}")
    return root


def fold_constants(root: Node) -> Node:
    if root.node_type == NodeType.LITERAL:
        # if we're already a literal (constant), we can't fold
        return root

    if root.node_type in {NodeType.COMPARISON_OPERATOR, NodeType.BINARY_OPERATOR}:
        # if we have a binary expression, try to fold each side
        root.left = fold_constants(root.left)
        root.right = fold_constants(root.right)

        # some expressions we can simplify to x or 0.
        if root.node_type == NodeType.BINARY_OPERATOR:
            if (
                root.value == "Multiply"
                and root.left.node_type == NodeType.LITERAL
                and root.right.node_type == NodeType.IDENTIFIER
                and root.left.value == 0
            ):
                # 0 * anything = 0
                root.left.schema_column = root.schema_column
                return root.left  # 0
            if (
                root.value == "Multiply"
                and root.right.node_type == NodeType.LITERAL
                and root.left.node_type == NodeType.IDENTIFIER
                and root.right.value == 0
            ):
                # anything * 0 = 0
                root.right.schema_column = root.schema_column
                return root.right  # 0
            if (
                root.value == "Multiply"
                and root.left.node_type == NodeType.LITERAL
                and root.right.node_type == NodeType.IDENTIFIER
                and root.left.value == 1
            ):
                # 1 * anything = anything
                root.right.query_column = root.query_column
                return root.right  # anything
            if (
                root.value == "Multiply"
                and root.right.node_type == NodeType.LITERAL
                and root.left.node_type == NodeType.IDENTIFIER
                and root.right.value == 1
            ):
                # anything * 1 = anything
                root.left.query_column = root.query_column
                return root.left  # anything
            if (
                root.value in "Plus"
                and root.left.node_type == NodeType.LITERAL
                and root.right.node_type == NodeType.IDENTIFIER
                and root.left.value == 0
            ):
                # 0 + anything = anything
                root.right.query_column = root.query_column
                return root.right  # anything
            if (
                root.value in ("Plus", "Minus")
                and root.right.node_type == NodeType.LITERAL
                and root.left.node_type == NodeType.IDENTIFIER
                and root.right.value == 0
            ):
                # anything +/- 0 = anything
                root.left.query_column = root.query_column
                return root.left  # anything
            if (
                root.value == "Divide"
                and root.right.node_type == NodeType.LITERAL
                and root.left.node_type == NodeType.IDENTIFIER
                and root.right.value == 1
            ):
                # anything / 1 = anything
                root.left.schema_column = root.schema_column
                return root.left  # anything
            if (
                root.value == "Divide"
                and root.right.node_type == NodeType.IDENTIFIER
                and root.left.node_type == NodeType.IDENTIFIER
                and root.right.schema_column.identity == root.left.schema_column.identity
            ):
                # anything / itself = 1 (0 is an exception)
                node = build_literal_node(1, root, OrsoTypes.INTEGER)
                node.schema_column = root.schema_column
                return node

    if root.node_type in {NodeType.AND, NodeType.OR, NodeType.XOR}:
        # try to fold each side of logical operators
        root.left = fold_constants(root.left)
        root.right = fold_constants(root.right)

        # If we have a logical expression and one side is a constant,
        # we can simplify further
        if root.node_type == NodeType.OR:
            if (
                root.left.node_type == NodeType.LITERAL
                and root.left.type == OrsoTypes.BOOLEAN
                and root.left.value
            ):
                # True OR anything is True
                root.left.schema_column = root.schema_column
                return root.left
            if (
                root.right.node_type == NodeType.LITERAL
                and root.right.type == OrsoTypes.BOOLEAN
                and root.right.value
            ):
                # anything OR True is True
                root.right.schema_column = root.schema_column
                return root.right
            if (
                root.left.node_type == NodeType.LITERAL
                and root.left.type == OrsoTypes.BOOLEAN
                and not root.left.value
            ):
                # False OR anything is anything
                root.right.schema_column = root.schema_column
                return root.right
            if (
                root.right.node_type == NodeType.LITERAL
                and root.right.type == OrsoTypes.BOOLEAN
                and not root.right.value
            ):
                # anything OR False is anything
                root.left.schema_column = root.schema_column
                return root.left

        elif root.node_type == NodeType.AND:
            if (
                root.left.node_type == NodeType.LITERAL
                and root.left.type == OrsoTypes.BOOLEAN
                and not root.left.value
            ):
                # False AND anything is False
                root.left.schema_column = root.schema_column
                return root.left
            if (
                root.right.node_type == NodeType.LITERAL
                and root.right.type == OrsoTypes.BOOLEAN
                and not root.right.value
            ):
                # anything AND False is False
                root.right.schema_column = root.schema_column
                return root.right
            if (
                root.left.node_type == NodeType.LITERAL
                and root.left.type == OrsoTypes.BOOLEAN
                and root.left.value
            ):
                # True AND anything is anything
                root.right.schema_column = root.schema_column
                return root.right
            if (
                root.right.node_type == NodeType.LITERAL
                and root.right.type == OrsoTypes.BOOLEAN
                and root.right.value
            ):
                # anything AND True is anything
                root.left.schema_column = root.schema_column
                return root.left

        return root

    identifiers = get_all_nodes_of_type(root, (NodeType.IDENTIFIER, NodeType.WILDCARD))
    functions = get_all_nodes_of_type(root, (NodeType.FUNCTION,))
    aggregators = get_all_nodes_of_type(root, (NodeType.AGGREGATOR,))

    if any(func.value in {"RANDOM", "RAND", "NORMAL", "RANDOM_STRING"} for func in functions):
        # Although they have no params, these are evaluated per row
        return root

    if len(identifiers) == 0 and len(aggregators) == 0:
        table = no_table_data.read()
        try:
            result = evaluate(root, table)[0]
            return build_literal_node(result, root, root.schema_column.type)
        except (ValueError, TypeError, KeyError) as err:  # nosec
            if not err:
                pass
            # what ever the reason, just skip
            # DEBUG:log (err)
    return root


class ConstantFoldingStrategy(OptimizationStrategy):
    def visit(self, node: LogicalPlanNode, context: OptimizerContext) -> OptimizerContext:
        """
        Constant Folding is when we precalculate expressions (or sub expressions)
        which contain only constant or literal values.
        """
        if not context.optimized_plan:
            context.optimized_plan = context.pre_optimized_tree.copy()  # type: ignore

        # fold constants when referenced in filter clauses (WHERE/HAVING)
        if node.node_type == LogicalPlanStepType.Filter:
            node.condition = fold_constants(node.condition)
            if node.condition.node_type == NodeType.LITERAL and node.condition.value:
                context.optimized_plan.remove_node(context.node_id, heal=True)
            else:
                context.optimized_plan[context.node_id] = node
        # fold constants when referenced in the SELECT clause
        if node.node_type == LogicalPlanStepType.Project:
            node.columns = [fold_constants(c) for c in node.columns]
            context.optimized_plan[context.node_id] = node

        return context

    def complete(self, plan: LogicalPlan, context: OptimizerContext) -> LogicalPlan:
        # No finalization needed for this strategy
        return plan



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Ordering of adjacent predicates using brute-force.

This is only used for small numbers of adjacent predicates - up to four. That is only 16 variations to test.
"""

import itertools


def calculate_predicate_costs(predicate_selectivity, predicate_execution_time):
    # Generate all possible predicate arrangements
    predicate_arrangements = list(itertools.permutations(range(len(predicate_selectivity))))

    # Evaluate execution time for each arrangement considering selectivity
    arrangement_costs = {}
    for arrangement in predicate_arrangements:
        cumulative_data_size = 1.0  # Assume initial data size is 1.0 (100%)
        execution_time = 0.0
        for i in arrangement:
            execution_time += predicate_execution_time[i] * cumulative_data_size
            cumulative_data_size *= predicate_selectivity[
                i
            ]  # Reduce data size based on selectivity
        arrangement_costs[arrangement] = execution_time

    return arrangement_costs


def print_arrangement_costs(arrangement_costs):
    for arrangement, cost in arrangement_costs.items():
        print(f"Arrangement {arrangement}: Execution Time = {cost}")


# Sample predicate selectivity and execution time estimates
predicate_selectivity = [0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5]
predicate_execution_time = [0.8, 0.2, 0.5, 0.1, 0.1, 0.4, 0.7]

# Calculate the costs for each arrangement
arrangement_costs = calculate_predicate_costs(predicate_selectivity, predicate_execution_time)

# Print the costs
print_arrangement_costs(arrangement_costs)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This is the async wrapper around the memory pool, it is used by the
async blob readers to move data to the synchronous code.
"""

import asyncio

from opteryx.compiled.structures import MemoryPool


class AsyncMemoryPool:
    def __init__(self, pool: MemoryPool):
        self.pool: MemoryPool = pool
        self.lock = asyncio.Lock()

    async def commit(self, data: bytes) -> int:
        async with self.lock:
            return self.pool.commit(data)

    async def read(self, ref_id: int) -> bytes:
        """
        In an async environment, we much more certain the bytes will be overwritten
        if we don't materialize them so we always create a copy.
        """
        async with self.lock:
            return self.pool.read(ref_id, zero_copy=False)

    async def release(self, ref_id: int):
        async with self.lock:
            self.pool.release(ref_id)

    def size(self):
        return self.pool.size



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
~~~
                      ┌───────────┐
                      │   USER    │
         ┌────────────┤           ◄────────────┐
         │SQL         └───────────┘            │
  ───────┼─────────────────────────────────────┼──────
         │                                     │
   ┌─────▼─────┐                               │
   │ SQL       │                               │
   │ Rewriter  │                               │
   └─────┬─────┘                               │
         │SQL                                  │Results
   ┌─────▼─────┐                         ┌─────┴─────┐
   │           │                         │           │
   │ Parser    │                         │ Executor  │
   └─────┬─────┘                         └─────▲─────┘
         │AST                                  │Plan
   ┌─────▼─────┐      ┌───────────┐      ┌─────┴─────┐
   │ AST       │      │           │      │ Physical  │
   │ Rewriter  │      │ Catalogue │      │ Planner   │
   └─────┬─────┘      └───────────┘      └─────▲─────┘
         │AST               │Schemas           │Plan
   ┌─────▼─────┐      ┌─────▼─────┐      ┌─────┴─────┐
   │ Logical   │ Plan │           │ Plan │           │
   │   Planner ├──────► Binder    ├──────► Optimizer │
   └───────────┘      └───────────┘      └───────────┘

~~~

The SQL Rewriter does the following:
- strips comments
- normalizes whitespace
- temporal extraction (this is non-standard and not part of the parser)

This compensates for missing temporal table support in the SQL parser (sqlparser-rs).
This is relatively complex for what it appears to be doing - it needs to account for
a number of situations whilst being able to reconstruct the SQL query as the parser
would expect it.

For information on temporal tables see:
https://blog.devgenius.io/a-query-in-time-introduction-to-sql-server-temporal-tables-145ddb1355d9

This supports the following syntaxes:

- FOR <timestamp>
- FOR DATES BETWEEN <timestamp> AND <timestamp>
- FOR DATES IN <range>
- FOR DATES SINCE <timestamp>

"""

import datetime
import re
from typing import List
from typing import Tuple

from opteryx.exceptions import InvalidTemporalRangeFilterError
from opteryx.utils import dates

COLLECT_RELATION = [
    r"FROM",
    r"INNER\sJOIN",
    r"CROSS\sJOIN",
    r"LEFT\sJOIN",
    r"LEFT\sOUTER\sJOIN",
    r"RIGHT\sJOIN",
    r"RIGHT\sOUTER\sJOIN",
    r"FULL\sJOIN",
    r"FULL\sOUTER\sJOIN",
    r"RIGHT\sSEMI\sJOIN",
    r"RIGHT\sANTI\sJOIN",
    r"LEFT\sSEMI\sJOIN",
    r"LEFT\sANTI\sJOIN",
    r"JOIN",
    r"CREATE\sTABLE",
    r"ANALYZE\sTABLE",
]

COLLECT_TEMPORAL = [r"FOR"]

STOP_COLLECTING = [
    r"GROUP\sBY",
    r"HAVING",
    r"LIKE",
    r"LIMIT",
    r"OFFSET",
    r"ON",
    r"ORDER\sBY",
    r"SHOW",
    r"SELECT",
    r"WHERE",
    r"WITH",
    r"USING",
    r";",
    r",",
    r"UNION",
]

COLLECT_ALIAS = [r"AS"]

BOUNDARIES = [r"(", r")"]

SQL_PARTS = (
    COLLECT_RELATION
    + COLLECT_TEMPORAL
    + STOP_COLLECTING
    + COLLECT_ALIAS
    + [
        r"DATES\sIN\s\w+",
        r"DATES\sBETWEEN\s[^\r\n\t\f\v]AND\s[^\r\n\t\f\v]",
        r"DATES\sSINCE\s\w+",
    ]
)

COMBINE_WHITESPACE_REGEX = re.compile(r"\r\n\t\f\v+")

# states for the collection algorithm
WAITING: int = 1
RELATION: int = 4
TEMPORAL: int = 16
ALIAS: int = 64
FUNCTION_RELATION: int = 128


def sql_parts(string):
    """
    Split a SQL statement into clauses
    """
    keywords = re.compile(
        r"(\,|\(|\)|\;|\t|\n|"
        + r"|".join([r"\b" + i.replace(r" ", r"\s") + r"\b" for i in SQL_PARTS])
        + r")",
        re.IGNORECASE,
    )
    # Match ", ', b", b', `
    # We match b prefixes separately after the non-prefix versions
    quoted_strings = re.compile(r'("[^"]*"|\'[^\']*\'|\b[bB]"[^"]*"|\b[bB]\'[^\']*\'|`[^`]*`)')

    parts = []
    for part in quoted_strings.split(string):
        if part and part[-1] in ("'", '"', "`"):
            if part[0] in ("b", "B"):
                parts.append(f"blob({part[1:]})")
            else:
                parts.append(part)
        else:
            for subpart in keywords.split(part):
                subpart = subpart.strip()
                if subpart:
                    parts.append(subpart)

    return parts


def parse_range(fixed_range):  # pragma: no cover
    fixed_range = fixed_range.upper()
    now = datetime.datetime.utcnow().replace(minute=0, second=0, microsecond=0)

    if fixed_range in ("PREVIOUS_MONTH", "LAST_MONTH"):
        # end the day before the first of this month
        end = now.replace(day=1) - datetime.timedelta(days=1)
        # start the first day of that month
        start = end.replace(day=1, hour=0, minute=0, second=0, microsecond=0)
    elif fixed_range == "THIS_MONTH":
        # start the first day of this month
        start = now.replace(day=1, hour=0, minute=0, second=0, microsecond=0)
        # end today
        end = now
    else:
        if parse_date(fixed_range):
            raise InvalidTemporalRangeFilterError(
                f"`THIS_MONTH`, `LAST_MONTH` expected, got `{fixed_range}`"
            )
        raise InvalidTemporalRangeFilterError(f"Unknown temporal range `{fixed_range}`")

    return start, end


def parse_date(date, end: bool = False):  # pragma: no cover
    if not date:
        return None

    now = datetime.datetime.utcnow().replace(minute=0, second=0, microsecond=0)

    if date == "TODAY":
        return now
    if date == "YESTERDAY":
        return (now - datetime.timedelta(days=1)).replace(hour=0)

    weekdays = [
        "MONDAY",
        "TUESDAY",
        "WEDNESDAY",
        "THURSDAY",
        "FRIDAY",
        "SATURDAY",
        "SUNDAY",
    ]
    if date in weekdays:
        # Find the weekday number (0=Monday, 1=Tuesday, ..., 6=Sunday)
        target_weekday = weekdays.index(date)
        # Find the current weekday number
        current_weekday = now.weekday()

        # Calculate how many days to subtract to get the last occurrence of the target weekday
        days_to_subtract = (current_weekday - target_weekday) % 7
        if days_to_subtract == 0:
            # If today is the target weekday, adjust to get the last week's same day
            days_to_subtract = 7

        # Calculate the most recent date for the target weekday
        most_recent_day = (now - datetime.timedelta(days=days_to_subtract)).replace(
            hour=0, minute=0, second=0, microsecond=0
        )
        return most_recent_day

    if date[0] == date[-1] and date[0] in ("'", '"', "`"):
        date = date[1:-1]

    parsed = dates.parse_iso(date)
    if parsed and end and len(date) <= 12:
        return parsed.replace(hour=23, minute=59)

    return parsed


def _temporal_extration_state_machine(
    parts: List[str],
) -> Tuple[List[Tuple[str, str]], str]:
    """
    Utilizes a state machine to extract the temporal information from the query
    and maintain the relation to filter information.

    Parameters:
        parts: List[str]
            SQL statement parts

    Returns:
        Tuple containing two lists, first with the temporal filters, second with the remaining SQL parts.
    """
    # We use a four state machine to extract the temporal information from the query
    # and maintain the relation to filter information.
    #
    # We separate out the two key parts of the algorithm, first we determine the state,
    # then we work out if the state transition means we should do something.
    #
    # We're essentially using a bit mask to record state and transitions.

    state = WAITING
    relation = ""
    temporal = ""
    query_collector = []
    temporal_range_collector = []
    open_count = 0
    for part in parts:
        # record the current state
        transition = [state]
        comparable_part = part.upper().replace(" ", r"\s")

        # work out what our current state is
        if comparable_part in BOUNDARIES:
            if relation == "":
                state = WAITING
            else:
                # function relations, like FAKE(234,234) need the items between the
                # brackets be be consumed
                state = FUNCTION_RELATION
                if comparable_part == "(":
                    open_count += 1
                if comparable_part == ")":
                    open_count -= 1
        if comparable_part in STOP_COLLECTING:
            if state == FUNCTION_RELATION and open_count > 0:
                pass
            else:
                state = WAITING
        if comparable_part in COLLECT_RELATION:
            state = RELATION
        if comparable_part in COLLECT_TEMPORAL:
            state = TEMPORAL
        if comparable_part in COLLECT_ALIAS:
            state = ALIAS
        transition.append(state)

        # based on what the state was and what it is now, do something
        if transition == [TEMPORAL, TEMPORAL]:
            temporal = (temporal + " " + part).strip()
        elif (
            transition
            in (
                [WAITING, WAITING],
                [TEMPORAL, RELATION],
                [RELATION, RELATION],
                [RELATION, WAITING],
                [ALIAS, RELATION],
                [ALIAS, WAITING],  # probably
                [FUNCTION_RELATION, RELATION],
            )
            and relation
        ):
            temporal_range_collector.append((relation, temporal))
            relation = ""
            temporal = ""
            if comparable_part == ",":
                state = RELATION
        elif transition == [RELATION, RELATION]:
            relation = part
        elif transition == [WAITING, TEMPORAL]:
            raise InvalidTemporalRangeFilterError(
                "Temporal `FOR` statements must directly follow the dataset they apply to."
            )
        else:
            pass

        if state != TEMPORAL:
            query_collector.append(part)

    # if we're at the end of we have a relation, emit it
    if relation:
        temporal_range_collector.append((relation, temporal))

    return temporal_range_collector, " ".join(query_collector)


def extract_temporal_filters(sql: str):  # pragma: no cover
    import shlex

    parts = sql_parts(sql)

    # extract the raw temporal information
    initial_collector, sql = _temporal_extration_state_machine(parts)

    final_collector = []

    for relation, for_date_string in initial_collector:
        start_date = None
        end_date = None

        for_date_string = for_date_string.upper()
        for_date = parse_date(for_date_string)

        if for_date:
            start_date = for_date
            end_date = for_date
            if for_date_string in ("TODAY", "YESTERDAY") or len(for_date_string) <= 12:
                start_date = start_date.replace(hour=0)
                end_date = end_date.replace(hour=23, minute=59)

        elif for_date_string.startswith("DATES BETWEEN "):
            parts = shlex.split(for_date_string)

            if len(parts) != 5:
                raise InvalidTemporalRangeFilterError(
                    "Invalid temporal range, expected format `FOR DATES BETWEEN <start> AND <end>`."
                )
            if parts[3] != "AND":
                raise InvalidTemporalRangeFilterError(
                    f"Invalid temporal range, expected `AND`, found `{parts[3]}`."
                )

            start_date = parse_date(parts[2])
            end_date = parse_date(parts[4], end=True)

            if start_date is None:
                raise InvalidTemporalRangeFilterError(
                    f"Invalid temporal range, expected a date for start of range, found `{parts[2]}`."
                )
            if end_date is None:
                raise InvalidTemporalRangeFilterError(
                    f"Invalid temporal range, expected a date for end of range, found `{parts[4]}`."
                )
            if start_date > end_date:
                raise InvalidTemporalRangeFilterError(
                    "Invalid temporal range, start of range is after end of range."
                )

        elif for_date_string.startswith("DATES IN "):
            parts = shlex.split(for_date_string)
            start_date, end_date = parse_range(parts[2])

        elif for_date_string.startswith("DATES SINCE "):
            parts = shlex.split(for_date_string)
            start_date = parse_date(parts[2])
            end_date = datetime.datetime.utcnow().replace(minute=0, second=0, microsecond=0)

        elif for_date_string:
            raise InvalidTemporalRangeFilterError(
                f"Unable to interpret temporal filter `{for_date_string}`"
            )

        final_collector.append(
            (
                relation,
                start_date,
                end_date,
            )
        )

    # we've rewritten the sql so make it sqlparser-rs compatible
    return sql, final_collector


def do_sql_rewrite(statement):
    return extract_temporal_filters(statement)



from .cython_list_ops import array_encode_utf8
from .cython_list_ops import cython_allop_eq
from .cython_list_ops import cython_allop_neq
from .cython_list_ops import cython_anyop_eq
from .cython_list_ops import cython_anyop_gt
from .cython_list_ops import cython_anyop_gte
from .cython_list_ops import cython_anyop_lt
from .cython_list_ops import cython_anyop_lte
from .cython_list_ops import cython_anyop_neq
from .cython_list_ops import cython_arrow_op
from .cython_list_ops import cython_get_element_op
from .cython_list_ops import cython_long_arrow_op



import fnmatch
from typing import Dict
from typing import List

import orjson


def load_permissions() -> List[Dict]:
    """
    Load permissions from a JSON file and return a list of permissions.

    If the file is not found, or an error occurs, return a default set of permissions.

    Returns:
        List[Dict]: A list of dictionaries where each dictionary represents a permission.
    """
    try:
        with open("permissions.json", "r", encoding="UTF8") as file:
            # Load each line as a JSON object and append a default permission entry.
            _permissions = [orjson.loads(line) for line in file] + [
                {"role": "opteryx", "permission": "READ", "table": "*"}
            ]
        return _permissions
    except FileNotFoundError:
        # Return a default permission if the file does not exist.
        return [{"role": "opteryx", "permission": "READ", "table": "*"}]
    except Exception as err:
        # Log the error and return a default permission in case of any other exceptions.
        print(f"[OPTERYX] Failed to load permissions: {err}")
        return [{"role": "opteryx", "permission": "READ", "table": "*"}]


# Load permissions once and make them globally accessible.
PERMISSIONS: List[Dict] = load_permissions()


def can_read_table(roles: List[str], table: str, action: str = "READ") -> bool:
    """
    Check if any of the provided roles have READ permission for the specified table.

    Parameters:
        roles (List[str]): A list of roles to check against permissions.
        table (str): The name of the table to check access for.

    Returns:
        bool: True if any role has READ permission for the table, False otherwise.
    """
    # If no permissions are loaded, default to allowing all reads.
    if not PERMISSIONS:
        return True

    for entry in PERMISSIONS:
        # Check if the permission, the role is in the provided roles,
        # and the table matches the pattern defined in the permission.
        if (
            entry["permission"] == action
            and entry["role"] in roles
            and fnmatch.fnmatch(table, entry["table"])
        ):
            return True

    # If no matching permission is found, deny access.
    return False



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from typing import List
from typing import Set
from typing import Tuple

from orso.schema import FlatColumn
from orso.schema import RelationSchema
from orso.tools import random_string
from orso.types import OrsoTypes

from opteryx.exceptions import AmbiguousDatasetError
from opteryx.exceptions import InvalidFunctionParameterError
from opteryx.exceptions import UnsupportedSyntaxError
from opteryx.managers.expression import NodeType
from opteryx.managers.expression import get_all_nodes_of_type
from opteryx.models import LogicalColumn
from opteryx.models import Node
from opteryx.planner.binder.binder import inner_binder
from opteryx.planner.binder.binder import locate_identifier_in_loaded_schemas
from opteryx.planner.binder.binder import merge_schemas
from opteryx.planner.binder.binding_context import BindingContext
from opteryx.planner.logical_planner import LogicalPlan
from opteryx.virtual_datasets import derived

CAMEL_TO_SNAKE = re.compile(r"(?<!^)(?=[A-Z])")


def get_mismatched_condition_column_types(node: Node, relaxed: bool = False) -> dict:
    """
    Checks that the types of the fields involved a comparison are the same on both sides.

    Parameters:
        node: Node
            The condition node representing the condition.

    Returns:
        a dictionary describing the columns
    """
    if node.node_type in (NodeType.AND, NodeType.OR, NodeType.XOR):
        left_mismatches = get_mismatched_condition_column_types(node.left, relaxed)
        right_mismatches = get_mismatched_condition_column_types(node.right, relaxed)
        return left_mismatches or right_mismatches

    elif node.node_type == NodeType.COMPARISON_OPERATOR:
        if node.value in (
            "InList",
            "NotInList",
            "Arrow",
            "LongArrow",
            "AtQuestion",
        ) or node.value.startswith(("AllOp", "AnyOp")):
            return None  # Some ops are meant to have different types
        left_type = node.left.schema_column.type if node.left.schema_column else None
        right_type = node.right.schema_column.type if node.right.schema_column else None

        if left_type and right_type and left_type != right_type:
            if (
                relaxed
                and (left_type.is_numeric() and right_type.is_numeric())
                or (left_type.is_temporal() and right_type.is_temporal())
                or (left_type.is_large_object() and right_type.is_large_object())
                or (left_type == 0 or right_type == 0)
            ):
                return None
            if left_type == OrsoTypes.NULL or right_type == OrsoTypes.NULL:
                return None  # None comparisons are allowed
            if (
                node.left.node_type == NodeType.COMPARISON_OPERATOR
                or node.right.node_type == NodeType.COMPARISON_OPERATOR
                or node.left.node_type == NodeType.BINARY_OPERATOR
                or node.right.node_type == NodeType.BINARY_OPERATOR
            ):
                return None  # it's compound so don't make a decision here
            return {
                "left_column": f"{node.left.source}.{node.left.value}",
                "left_type": left_type.name,
                "left_node": node.left,
                "right_column": f"{node.right.source}.{node.right.value}",
                "right_type": right_type.name,
                "right_node": node.right,
            }

    return None  # if we reach here, it means we didn't find any inconsistencies


def extract_join_fields(
    condition_node: Node,
    left_relation_names: List[str],
    right_relation_names: List[str],
) -> Tuple[List[str], List[str]]:
    """
    Extracts join fields from a condition node that may have multiple ANDed conditions.

    Parameters:
        condition_node: Node
            The condition node in the join clause.
        left_relation_name: str
            Name of the left relation.
        right_relation_name: str
            Name of the right relation.

    Returns:
        Tuple[List[str], List[str]]
            Lists of columns participating in the join from the left and right tables.
    """
    left_fields = []
    right_fields = []

    if condition_node.node_type == NodeType.AND:
        left_fields_1, right_fields_1 = extract_join_fields(
            condition_node.left, left_relation_names, right_relation_names
        )
        left_fields_2, right_fields_2 = extract_join_fields(
            condition_node.right, left_relation_names, right_relation_names
        )

        left_fields.extend(left_fields_1)
        left_fields.extend(left_fields_2)

        right_fields.extend(right_fields_1)
        right_fields.extend(right_fields_2)

    elif condition_node.node_type == NodeType.COMPARISON_OPERATOR and condition_node.value == "Eq":
        if any(
            [
                condition_node.left.node_type not in (NodeType.IDENTIFIER, NodeType.LITERAL),
                condition_node.right.node_type not in (NodeType.IDENTIFIER, NodeType.LITERAL),
            ]
        ):
            raise UnsupportedSyntaxError("JOIN conditions only support column comparisons.")
        if (
            condition_node.left.source in left_relation_names
            and condition_node.right.source in right_relation_names
        ):
            left_fields.append(condition_node.left.schema_column.identity)
            right_fields.append(condition_node.right.schema_column.identity)
        elif (
            condition_node.left.source in right_relation_names
            and condition_node.right.source in left_relation_names
        ):
            right_fields.append(condition_node.left.schema_column.identity)
            left_fields.append(condition_node.right.schema_column.identity)

    return left_fields, right_fields


def convert_using_to_on(
    using_fields: Set[str],
    left_relation_names: List[str],
    right_relation_names: List[str],
) -> Node:
    """
    Converts a USING field to an ON field for JOIN operations.

    Parameters:
        using_fields: Set[str]
            Set of common fields to use for joining.
        left_relation_names: List[str]
            Names of the left relations.
        right_relation_names: List[str]
            Names of the right relations.

    Returns:
        Node
            The condition node representing the ON clause.
    """
    all_conditions = []

    # Loop through all combinations of left and right relation names
    for left_relation_name in left_relation_names:
        for right_relation_name in right_relation_names:
            conditions = []
            for field in using_fields:
                condition = Node(
                    node_type=NodeType.COMPARISON_OPERATOR,
                    value="Eq",
                    do_not_create_column=True,
                )
                condition.left = LogicalColumn(
                    node_type=NodeType.IDENTIFIER,
                    source=left_relation_name,
                    source_column=field,
                )
                condition.right = LogicalColumn(
                    node_type=NodeType.IDENTIFIER,
                    source=right_relation_name,
                    source_column=field,
                )
                conditions.append(condition)

            if len(conditions) == 1:
                all_conditions.append(conditions[0])
            else:
                # Create a tree of ANDed conditions
                while len(conditions) > 1:
                    new_conditions = []
                    for i in range(0, len(conditions), 2):
                        if i + 1 < len(conditions):
                            and_node = Node(node_type=NodeType.AND, do_not_create_column=True)
                            and_node.left = conditions[i]
                            and_node.right = conditions[i + 1]
                            new_conditions.append(and_node)
                        else:
                            new_conditions.append(conditions[i])
                    conditions = new_conditions
                all_conditions.append(conditions[0])

    return conditions[0]


class BinderVisitor:
    """
    The BinderVisitor visits each node in the query plan and adds catalogue information
    to each node. This includes:

    - identifiers, bound from the schemas
    - functions and aggregatros, bound from the function catalogue
    - variables, bound from the variables collection

    """

    def visit_node(self, node: Node, context: BindingContext) -> Tuple[Node, BindingContext]:
        """
        Visits a given node and returns a new node and context after binding catalog information.

        Parameters:
            node: Node
                The query plan node to visit.
            context: Dict[str, Any]
                The current binding context.

        Returns:
            Tuple[Node, Dict]
            The node and context after binding.
        """
        node_type = node.node_type.name  # type:ignore
        visit_method_name = f"visit_{CAMEL_TO_SNAKE.sub('_', node_type).lower()}"
        visit_method = getattr(self, visit_method_name, None)
        if visit_method is None:
            return node, context

        return_node, return_context = visit_method(node.copy(), context.copy())

        # DEBUG: from opteryx.exceptions import InvalidInternalStateError
        # DEBUG:
        # DEBUG: if not isinstance(return_context, BindingContext):
        # DEBUG:     raise InvalidInternalStateError(
        # DEBUG:         f"Internal Error - function '{visit_method_name}' didn't return a BindingContext"
        # DEBUG:     )
        # DEBUG:
        # DEBUG: if not all(isinstance(schema, RelationSchema) for schema in context.schemas.values()):
        # DEBUG:     raise InvalidInternalStateError(
        # DEBUG:         f"Internal Error - function '{visit_method_name}' returned invalid Schemas"
        # DEBUG:     )
        # DEBUG:
        # DEBUG: if not all(isinstance(col, (Node, LogicalColumn)) for col in return_node.columns or []):
        # DEBUG:     raise InvalidInternalStateError(
        # DEBUG:         f"Internal Error - function '{visit_method_name}' put unexpected items in 'columns' attribute"
        # DEBUG:     )
        # DEBUG:
        # DEBUG: if return_node.node_type.name != "Scan" and return_node.columns is None:
        # DEBUG:     raise InvalidInternalStateError(
        # DEBUG:         f"Internal Error - function {visit_method_name} did not populate 'columns'"
        # DEBUG:     )

        return return_node, return_context

    def visit_aggregate_and_group(
        self, node: Node, context: BindingContext
    ) -> Tuple[Node, BindingContext]:
        """
        Handles the binding logic for aggregate and group nodes.

        This function maps the field to the existing schema fields, disposes of the existing
        schemas, and replaces it with a new '$group-by' schema.

        Parameters:
            node: Node
                The node containing the aggregate and group data.
            context: Optional[Dict[str, Any]]
                The current binding context, defaults to None.

        Returns:
            Tuple[Node, Dict[str, Any]]
            The modified node and the updated context.
        """
        tmp_aggregates: tuple = tuple()
        if node.aggregates:
            tmp_aggregates, _ = zip(
                *(inner_binder(aggregate, context) for aggregate in node.aggregates)
            )
            node.aggregates = list(tmp_aggregates)

        # We're going to trim down the schemas to just the columns used in the GROUP BY.
        # 1) the easy one - the columns explictly in the GROUP BY
        columns_to_keep = set()
        if node.groups:
            tmp_groups, _ = zip(*(inner_binder(group, context) for group in node.groups))
            columns_to_keep = {col.schema_column.identity for col in tmp_groups}
        # remove literals in the GROUP BY clause, they form one group
        node.groups = [g for g in node.groups if g.node_type != NodeType.LITERAL]
        # 2) the columns referenced in the SELECT
        node.columns = get_all_nodes_of_type(
            node.aggregates + node.groups, select_nodes=(NodeType.IDENTIFIER,)
        )
        all_identifiers = [node.schema_column.identity for node in node.columns]
        columns_to_keep = columns_to_keep.union(all_identifiers)

        for name, schema in list(context.schemas.items()):
            schema_columns = [
                column for column in schema.columns if column.identity in columns_to_keep
            ]
            if schema_columns:
                context.schemas[name].columns = schema_columns
            else:
                context.schemas.pop(name)

        for array_agg in [agg for agg in tmp_aggregates if agg.value == "ARRAY_AGG"]:
            if not node.groups:
                raise UnsupportedSyntaxError(
                    "ARRAY_AGG requires a GROUP BY clause, and cannot GROUP BY a literal value."
                )
            if array_agg.order:
                if len(array_agg.order) > 1:
                    raise UnsupportedSyntaxError(
                        "ARRAY_AGG can only ORDER BY the aggregated column."
                    )
                if array_agg.order[0][0].current_name != array_agg.parameters[0].current_name:
                    raise UnsupportedSyntaxError(
                        "ARRAY_AGG can only ORDER BY the aggregated column."
                    )

        # we should always have a derived schema
        if "$derived" not in context.schemas:
            context.schemas["$derived"] = derived.schema()

        # the aggregates and any calculated expressions in the SELECT should be in $derived
        context.schemas["$derived"].columns.extend(col.schema_column for col in node.aggregates)
        node.schema = context.schemas["$derived"]
        return node, context

    visit_aggregate = visit_aggregate_and_group

    def visit_distinct(self, node: Node, context: BindingContext) -> Tuple[Node, BindingContext]:
        node.columns = []
        if node.on:
            # Bind the local columns to physical columns
            node.on, group_contexts = zip(*(inner_binder(col, context) for col in node.on))
            context.schemas = merge_schemas(*[ctx.schemas for ctx in group_contexts])
            node.columns = get_all_nodes_of_type(node.on, (NodeType.IDENTIFIER,))

        return node, context

    def visit_exit(self, node: Node, context: BindingContext) -> Tuple[Node, BindingContext]:
        # clear the derived schema
        context.schemas.pop("$derived", None)
        context.schemas["$derived"] = derived.schema()

        seen = set()
        needs_qualifier = any(
            column.name in seen or seen.add(column.name) is not None  # type: ignore
            for schema in context.schemas.values()
            for column in schema.columns
        )

        def name_column(qualifier, column):
            for projection_column in node.columns:
                if (
                    projection_column.schema_column
                    and projection_column.schema_column.identity == column.identity
                ):
                    if projection_column.alias:
                        return projection_column.alias

                    if len(context.relations) > 1 or needs_qualifier:
                        if isinstance(projection_column, LogicalColumn):
                            if qualifier:
                                projection_column.source = qualifier
                            return projection_column.qualified_name
                        return f"{qualifier}.{column.name}"

                    if projection_column.query_column:
                        return str(projection_column.query_column)
                    if projection_column.current_name:
                        return projection_column.current_name

            return column.name

        def keep_column(column, identities):
            if len(node.columns) == 1 and node.columns[0].node_type == NodeType.WILDCARD:
                if node.columns[0].value:
                    if isinstance(column.origin, str):
                        column.origin = [column.origin]
                    if node.columns[0].value[0] in column.origin:
                        identities.append(column.identity)
                        return True
                    else:
                        return False
                identities.append(column.identity)
                return True
            return column.identity in identities

        identities = []
        for column in (col for col in node.columns if col.node_type != NodeType.WILDCARD):
            new_col, _ = inner_binder(column, context)
            identities.append(new_col.schema_column.identity)

        columns = []
        for qualifier, schema in context.schemas.items():
            for column in schema.columns:
                if keep_column(column, identities):
                    column_name = name_column(qualifier=qualifier, column=column)
                    column_reference = LogicalColumn(
                        node_type=NodeType.IDENTIFIER,
                        source_column=column_name,
                        source=None,
                        alias=None,
                        schema_column=column,
                    )
                    columns.append(column_reference)

        # we bound as we came across items in schemas, not the order the user wants them
        desired_order = {id: index for index, id in enumerate(identities)}
        node.columns = sorted(columns, key=lambda item: desired_order[item.schema_column.identity])

        context.schemas["$derived"] = derived.schema()

        return node, context

    def visit_filter(self, node: Node, context: BindingContext) -> Tuple[Node, BindingContext]:
        # We don't update the context, otherwise we'd be adding the predicates as columns
        original_context = context.copy()
        node.condition, context = inner_binder(node.condition, context)
        node.columns = get_all_nodes_of_type(node.condition, (NodeType.IDENTIFIER,))
        node.relations = node.condition.relations or {}

        return node, original_context

    def visit_function_dataset(
        self, node: Node, context: BindingContext
    ) -> Tuple[Node, BindingContext]:
        # We need to build the schema and add it to the schema collection.
        if node.function == "VALUES":
            relation_name = node.alias or f"$values-{random_string()}"
            columns = [
                LogicalColumn(
                    node_type=NodeType.IDENTIFIER,
                    source_column=column,
                    source=relation_name,
                    schema_column=FlatColumn(name=column, type=0),
                )
                for column in node.columns
            ]
            schema = RelationSchema(
                name=relation_name,
                columns=[c.schema_column for c in columns],
            )
            context.schemas[relation_name] = schema
            node.columns = columns
            node.schema = schema
        elif node.function == "UNNEST":
            relation_name = node.alias

            columns = [
                LogicalColumn(
                    node_type=NodeType.IDENTIFIER,
                    source_column=node.unnest_target,
                    source=relation_name,
                    schema_column=FlatColumn(name=node.unnest_target, type=0),
                )
            ]
            schema = RelationSchema(name=relation_name, columns=[c.schema_column for c in columns])
            context.schemas[relation_name] = schema
            node.columns = columns
            node.schema = schema
        elif node.function == "GENERATE_SERIES":
            node.relation_name = node.alias
            columns = [
                LogicalColumn(
                    node_type=NodeType.IDENTIFIER,
                    source_column=node.alias,
                    source=node.relation_name,
                    schema_column=FlatColumn(name=node.alias, type=0),
                )
            ]
            schema = RelationSchema(
                name=node.relation_name,
                columns=[c.schema_column for c in columns],
            )
            context.schemas[node.relation_name] = schema
            node.columns = columns
            node.schema = schema
        elif node.function == "HTTP":
            node.relation_name = node.alias
            node.url = node.args[0].value

            import requests

            from opteryx.utils.file_decoders import get_decoder

            decoder = get_decoder(node.url)
            response = requests.get(node.url, timeout=60)

            response.raise_for_status()
            row_count, column_count, data = decoder(response.content, force_read=True)

            schema = RelationSchema(
                name=node.relation_name,
                columns=[FlatColumn.from_arrow(field) for field in data.schema],
            )

            context.schemas[node.relation_name] = schema

            node.data = data
            node.schema = schema
        elif node.function == "FAKE":
            from orso.schema import ColumnDisposition

            node.relation_name = node.alias
            node.rows = int(node.args[0].value)

            if len(node.args) < 2:
                raise InvalidFunctionParameterError(
                    "FAKE function expects at least two parameters, the number of rows, and then either the number of columns, or an array of the column types."
                )

            if node.args[1].node_type == NodeType.NESTED:
                column_definition = [node.args[1].centre]
            else:
                column_definition = node.args[1].value

            special_handling = {
                "NAME": (OrsoTypes.VARCHAR, ColumnDisposition.NAME),
                "AGE": (OrsoTypes.INTEGER, ColumnDisposition.AGE),
            }

            columns = []
            if isinstance(column_definition, tuple):
                for i, column_type in enumerate(column_definition):
                    name = node.columns[i] if i < len(node.columns) else f"column_{i}"
                    column_type = str(column_type).upper()
                    if column_type in special_handling:
                        actual_type, disposition = special_handling[column_type]
                        schema_column = FlatColumn(
                            name=name, type=actual_type, disposition=disposition
                        )
                    else:
                        schema_column = FlatColumn(name=name, type=column_type)
                    columns.append(
                        LogicalColumn(
                            node_type=NodeType.IDENTIFIER,
                            source_column=schema_column.name,
                            source=node.alias,
                            schema_column=schema_column,
                        )
                    )
                schema = RelationSchema(
                    name=node.alias,
                    columns=[c.schema_column for c in columns],
                )
                node.columns = columns
                node.schema = schema
            else:
                try:
                    column_definition = int(column_definition)  # type: ignore
                except TypeError:
                    raise InvalidFunctionParameterError(
                        "Expected number of rows for 'FAKE' function or list of column types. Are you missing parenthesis?"
                    )
                names = node.columns + tuple(
                    f"column_{i}"
                    for i in range(len(node.columns), column_definition)  # type: ignore
                )
                node.columns = [
                    LogicalColumn(
                        node_type=NodeType.IDENTIFIER,
                        source_column=names[i],
                        source=node.alias,
                        schema_column=FlatColumn(name=names[i], type=OrsoTypes.INTEGER),
                    )
                    for i in range(column_definition)  # type: ignore
                ]

            schema = RelationSchema(
                name=node.relation_name,
                columns=[c.schema_column for c in node.columns],
            )
            context.schemas[node.relation_name] = schema
            node.schema = schema
        else:
            raise UnsupportedSyntaxError(f"{node.function} cannot be used in place of a table.")
        return node, context

    def visit_join(self, node: Node, context: BindingContext) -> Tuple[Node, BindingContext]:
        """
        Visits a JOIN node and handles different types of joins.

        Parameters:
            node: Node
                The node representing the join operation.
            context: Dict
                The context containing relevant information like schemas.

        Returns:
            Tuple[Node, Dict]
                Updated node and context.
        """
        node.columns = []
        # Handle 'natural join' by converting to a 'using'
        if node.type == "natural join":
            left_columns = [
                col
                for relation_name in node.left_relation_names
                for col in context.schemas[relation_name].column_names
            ]
            right_columns = [
                col
                for relation_name in node.right_relation_names
                for col in context.schemas[relation_name].column_names
            ]
            node.using = [
                Node("temp", value=n) for n in set(left_columns).intersection(right_columns)
            ]
            node.type = "inner"
        # Handle 'using' by converting to a an 'on'
        if node.using:
            node.on = convert_using_to_on(
                {n.value for n in node.using},
                node.left_relation_names,
                node.right_relation_names,
            )
        if node.on:
            # All except CROSS JOINs have been mapped to have an ON condition
            # The JOIN operator only support ON conditions.
            comparisons = get_all_nodes_of_type(node.on, (NodeType.COMPARISON_OPERATOR,))
            if not all(com.value == "Eq" for com in comparisons):
                from opteryx.exceptions import UnsupportedSyntaxError

                raise UnsupportedSyntaxError("Only JOINs with equals comparisons supported")

            node.on, context = inner_binder(node.on, context)
            node.left_columns, node.right_columns = extract_join_fields(
                node.on, node.left_relation_names, node.right_relation_names
            )
            mismatches = get_mismatched_condition_column_types(node.on)
            if mismatches:
                from opteryx.exceptions import IncompatibleTypesError

                raise IncompatibleTypesError(**mismatches)

            # we need to put the referenced columns into the columns attribute for the
            # optimizers
            node.columns = get_all_nodes_of_type(node.on, (NodeType.IDENTIFIER,))

        if node.using:
            # Remove the columns used in the join condition from both relations, they're in
            # the result set but not belonging to either table, whilst still belonging to both.
            # We create a new schema to put them in, $shared-nnn.
            columns = []

            # Loop through all using fields in the node
            left_relation_name = ""
            right_relation_name = ""
            for column_name in (n.value for n in node.using):
                # Pop the column from the left relation
                for left_relation_name in node.left_relation_names:
                    left_column = context.schemas[left_relation_name].pop_column(column_name)

                # Pop the column from the right relation
                for right_relation_name in node.right_relation_names:
                    right_column = context.schemas[right_relation_name].pop_column(column_name)

                # we need to decide which column we're going to keep
                if node.type in ("right anti", "right semi"):
                    right_column.origin = [left_relation_name, right_relation_name]
                    columns.append(right_column)
                else:
                    left_column.origin = [left_relation_name, right_relation_name]
                    columns.append(left_column)

            # shared columns exist in both schemas in some uses and in neither in others
            context.schemas[f"$shared-{random_string()}"] = RelationSchema(
                name=f"^{left_relation_name}#^{right_relation_name}#", columns=columns
            )

        # SEMI and ANTI joins only return columns from one table
        if node.type in ("left anti", "left semi"):
            for schema in node.right_relation_names:
                context.schemas.pop(schema)
        if node.type in ("right anti", "right semi"):
            for schema in node.left_relation_names:
                context.schemas.pop(schema)

        # If we have an unnest_column, how how it is bound is different to other columns
        if node.unnest_column:
            # this is the column which is being unnested
            node.unnest_column, context = inner_binder(node.unnest_column, context)
            node.columns += [node.unnest_column]
            # this is the column that is being created - find it from its name
            node.unnest_target, found_source_relation = locate_identifier_in_loaded_schemas(
                node.unnest_alias, context.schemas
            )
            if node.unnest_column.schema_column.type not in (
                OrsoTypes._MISSING_TYPE,
                OrsoTypes.ARRAY,
                0,
            ):
                from opteryx.exceptions import IncorrectTypeError

                raise IncorrectTypeError(
                    f"CROSS JOIN UNNEST requires an ARRAY type column, not {node.unnest_column.schema_column.type}."
                )

        if node.type == "inner" and node.on is None:
            from opteryx.exceptions import SqlError

            raise SqlError("INNER and NATURAL joins must have a either an ON or USING condition.")

        return node, context

    def visit_order(self, node: Node, context: BindingContext) -> Tuple[Node, BindingContext]:
        order_by = []
        columns = []
        for column, direction in node.order_by:
            bound_column, context = inner_binder(column, context)

            order_by.append(
                (
                    bound_column,
                    "ascending" if direction else "descending",
                )
            )
            columns.append(bound_column)

        node.order_by = order_by
        node.columns = columns
        return node, context

    def visit_project(self, node: Node, context: BindingContext) -> Tuple[Node, BindingContext]:
        columns = []

        # Handle wildcards, including qualified wildcards.
        for column in node.columns + node.order_by_columns:
            if column.node_type != NodeType.WILDCARD:
                columns.append(column)
            else:
                # Handle qualified wildcards
                for name, schema in list(context.schemas.items()):
                    if (
                        name == column.value[0]
                        or name.startswith("$shared")
                        and f"^{column.value[0]}#" in schema.name
                    ):
                        for schema_column in schema.columns:
                            column_reference = LogicalColumn(
                                node_type=NodeType.IDENTIFIER,  # column type
                                source_column=schema_column.name,  # the source column
                                source=column.value[0],  # the source relation
                                schema_column=schema_column,
                            )
                            columns.append(column_reference)
                    if name.startswith("$shared") and f"^{column.value[0]}#" in schema.name:
                        context.schemas.pop(name)

                    context.schemas[column.value[0]] = RelationSchema(
                        name=name, columns=[col.schema_column for col in columns]
                    )

        node.columns = columns

        # Bind the local columns to physical columns
        node.columns, group_contexts = zip(*(inner_binder(col, context) for col in node.columns))
        context.schemas = merge_schemas(*[ctx.schemas for ctx in group_contexts])

        # Check for duplicates
        all_identities = [c.schema_column.identity for c in node.columns]
        if len(set(all_identities)) != len(all_identities):
            from collections import Counter

            from opteryx.exceptions import AmbiguousIdentifierError

            duplicates = [column for column, count in Counter(all_identities).items() if count > 1]
            matches = {c.value for c in node.columns if c.schema_column.identity in duplicates}
            raise AmbiguousIdentifierError(
                message=f"Query result contains multiple instances of the same column(s) - `{'`, `'.join(matches)}`"
            )

        # Remove columns not being projected from the schemas, and remove empty schemas
        columns = []
        for relation, schema in list(context.schemas.items()):
            schema_columns = [
                column for column in schema.columns if column.identity in all_identities
            ]
            if len(schema_columns) == 0:
                context.schemas.pop(relation)
            else:
                schema.columns = schema_columns
                for column in node.columns:
                    if column.schema_column.identity in [i.identity for i in schema_columns]:
                        columns.append(column)

        # We always have a $derived schema, even if it's empty
        if "$derived" in context.schemas:
            context.schemas["$project"] = context.schemas.pop("$derived")
            context.schemas["$project"].name = "$project"
        if "$derived" not in context.schemas:
            context.schemas["$derived"] = derived.schema()

        node.columns = columns

        return node, context

    def visit_scan(self, node: Node, context: BindingContext) -> Tuple[Node, BindingContext]:
        from opteryx.connectors import connector_factory
        from opteryx.connectors.capabilities import Asynchronous
        from opteryx.connectors.capabilities import Cacheable
        from opteryx.connectors.capabilities import Partitionable
        from opteryx.connectors.capabilities.cacheable import async_read_thru_cache
        from opteryx.connectors.capabilities.cacheable import read_thru_cache
        from opteryx.managers.catalog import catalog_factory
        from opteryx.managers.permissions import can_read_table
        from opteryx.managers.schemes.tarchia_schema import TarchiaScheme

        if node.alias in context.relations:
            raise AmbiguousDatasetError(dataset=node.alias)

        catalog_table = None
        node.found_in_catalog = False

        if node.relation.count(".") == 1:
            catalog_provider = catalog_factory()
            catalog_table = catalog_provider.table_exists(node.relation)
            if catalog_table:
                node.found_in_catalog = True
                node.disposition = catalog_table.get("disposition")
                node.location = catalog_table.get("location")

                if node.disposition == "EXTERNAL":
                    # explicitly told it's external
                    node.connector = connector_factory(node.location, statistics=context.statistics)
                else:
                    node.connector = connector_factory(
                        node.location, statistics=context.statistics, partition_scheme=TarchiaScheme
                    )
                    node.connector.dataset = node.relation

                catalog_schema = catalog_table.get("schema", {})
                catalog_schema["name"] = catalog_table.get("location")
                node.schema = RelationSchema.from_dict(catalog_schema)

        if not catalog_table:
            # work out which connector will be serving this request
            node.connector = connector_factory(node.relation, statistics=context.statistics)

        # ensure this user can read the table
        if not can_read_table(context.connection.memberships, node.relation):
            raise PermissionError(f"User does not have permission to read {node.relation}")

        connector_capabilities = node.connector.__class__.mro()

        if hasattr(node.connector, "variables"):
            node.connector.variables = context.connection.variables
        if Partitionable in connector_capabilities:
            node.connector.start_date = node.start_date
            node.connector.end_date = node.end_date
        if Cacheable in connector_capabilities:
            # We add the caching mechanism here if the connector is Cacheable and
            # we've not disable caching
            if "NO_CACHE" in (node.hints or []):
                pass
            if Asynchronous in connector_capabilities:
                original_read_blob = node.connector.async_read_blob
                node.connector.async_read_blob = async_read_thru_cache(original_read_blob)
            else:
                original_read_blob = node.connector.read_blob
                node.connector.read_blob = read_thru_cache(original_read_blob)

        if not node.found_in_catalog:
            node.schema = node.connector.get_dataset_schema()
        node.schema.aliases.append(node.alias)
        context.schemas[node.alias] = node.schema
        for column in node.schema.columns:
            column.origin = [node.alias]
            if catalog_table:
                column.origin.append(catalog_table["location"])
        context.relations[node.alias] = node.connector.__mode__

        return node, context

    def visit_set(self, node: Node, context: BindingContext) -> Tuple[Node, BindingContext]:
        node.variables = context.connection.variables
        node.columns = []
        return node, context

    def visit_show_columns(
        self, node: Node, context: BindingContext
    ) -> Tuple[Node, BindingContext]:
        node.schema = context.schemas[node.relation]
        node.columns = []
        for schema_column in node.schema.columns:
            column_reference = LogicalColumn(
                node_type=NodeType.IDENTIFIER,  # column type
                source_column=schema_column.name,  # the source column
                source=node.relation,  # the source relation
                schema_column=schema_column,
            )
            node.columns.append(column_reference)
        return node, context

    def visit_subquery(self, node: Node, context: BindingContext) -> Tuple[Node, BindingContext]:
        node, context = self.visit_exit(node, context)

        # we sack all the tables we previously knew and create a new set of schemas here
        columns: list = []
        source_relations: list = []
        for name, schema in context.schemas.items():
            for schema_column in schema.columns:
                # Find the column in the projection if it exists
                projection_column = next(
                    (
                        column
                        for column in node.columns
                        if column.schema_column.identity == schema_column.identity
                    ),
                    None,
                )
                if not schema_column.origin:
                    schema_column.origin = []
                source_relations.extend(schema_column.origin or [])
                projection_column.source = node.alias
                schema_column.origin += [node.alias]

                schema_column.name = (
                    projection_column.current_name if projection_column else schema_column.name
                )

                if "." in schema_column.name:
                    # If the column is not in the projection, it should retain its name without any prefix
                    schema_column.name = schema_column.name.split(".")[-1]

                schema_column.aliases = []
                columns.append(schema_column)
            if name[0] != "$" and name in context.relations:
                context.relations.pop(name)
        context.relations[node.alias] = "subquery"

        schema = RelationSchema(name=node.alias, columns=columns)

        context.schemas = {"$derived": derived.schema(), node.alias: schema}
        context.relations[node.alias] = "subquery"
        node.schema = schema
        node.source_relations = set(source_relations)
        return node, context

    def visit_union(self, node: Node, context: BindingContext) -> Tuple[Node, BindingContext]:
        for relation in node.right_relation_names:
            context.schemas.pop(relation, None)
        context.relations = {n: "union" for n in node.left_relation_names}

        if len(node.columns) == 1 and node.columns[0].node_type == NodeType.WILDCARD:
            columns = []
            for schema_name in node.left_relation_names:
                for schema_column in context.schemas[schema_name].columns:
                    columns.append(
                        LogicalColumn(
                            node_type=NodeType.IDENTIFIER,  # column type
                            source_column=schema_column.name,  # the source column
                            schema_column=schema_column,
                        )
                    )
            node.columns = columns

        node, context = self.visit_exit(node, context)
        return node, context

    def post_bind(self, node):
        # The binder skips calculated fields when it performs binding because
        # sometimes it doesn't have access to all of the fields used in the
        # calculation - so we bind these now
        seen: dict = {}

        def _inner(branch):
            if branch.fully_bound == False:
                if branch.schema_column.identity in seen:
                    branch = seen[branch.schema_column.identity]
            elif branch.schema_column:
                seen[branch.schema_column.identity] = branch.copy()
            for attr in ("left", "right", "centre"):
                if hasattr(branch, attr) and getattr(branch, attr) is not None:
                    setattr(branch, attr, _inner(getattr(branch, attr)))
            if branch.parameters:
                branch.parameters = [_inner(p) for p in branch.parameters]
            return branch

        if node.condition:
            node.condition = _inner(node.condition)
        if node.columns:
            # if it doesn't have a schema column here - we can remove it
            node.columns = [_inner(c) for c in node.columns if c.schema_column is not None]
        return node

    def traverse(
        self, graph: LogicalPlan, node: Node, context: BindingContext
    ) -> Tuple[LogicalPlan, BindingContext]:
        """
        Traverses the given graph starting at the given node and calling the
        appropriate visit methods for each node in the graph. This method uses
        a post-order traversal, which visits the children of a node before
        visiting the node itself.

        Args:
            graph: The graph to traverse.
            node: The node to start the traversal from.
            context: An optional context object to pass to each visit method.
        Returns:
            A tuple containing the updated graph and the context.
        """
        # Recursively visit children
        children = graph.ingoing_edges(node)

        if children:
            exit_context = context.copy()
            for child in children:
                # Each peer gets the exact copy of the context so they don't affect each other
                _, child_context = self.traverse(graph, child[0], context.copy())
                # Assuming merge_schemas is a function that merges the schemas from two contexts
                exit_context.schemas = merge_schemas(child_context.schemas, exit_context.schemas)

                # Update relations if necessary
                merged_relations = {
                    **context.relations,
                    **exit_context.relations,
                    **child_context.relations,
                }
                context.relations = merged_relations

            context.schemas = merge_schemas(context.schemas, exit_context.schemas)

        # Visit node and return updated context
        return_node, context = self.visit_node(graph[node], context=context)

        return_node.all_relations = {
            value for value in [return_node.relation, return_node.alias] if value is not None
        }
        sub_plan = {
            x for (s, t, r) in graph.breadth_first_search(node, reverse=True) for x in (s, t)
        }
        for plan_node_id in sub_plan:
            plan_node = graph[plan_node_id]
            if plan_node.alias:
                return_node.all_relations.add(plan_node.alias)
            if plan_node.relation:
                return_node.all_relations.add(plan_node.relation)
            if plan_node.all_relations:
                return_node.all_relations.update(plan_node.all_relations)

        return_node = self.post_bind(return_node)
        graph[node] = return_node
        return graph, context



import decimal

import pyarrow
from orso.types import OrsoTypes

import opteryx
from opteryx.managers.expression import NodeType
from opteryx.managers.expression import evaluate
from opteryx.managers.expression.ops import filter_operations
from opteryx.models import Node

OPERATORS = [
    "Eq",
    "NotEq",
    "Gt",
    "GtEq",
    "Lt",
    "LtEq",
    "Like",
    "ILike",
    "NotLike",
    "NotILike",
    "InList",
    "RLike",
    "NotRLike",
]

LITERALS = [
    (OrsoTypes.BOOLEAN, True),
    (OrsoTypes.ARRAY, ["a", "b", "c"]),
    (OrsoTypes.INTEGER, 0),
    (OrsoTypes.DOUBLE, 1e10),
    (OrsoTypes.DECIMAL, decimal.Decimal(4)),
    (OrsoTypes.INTEGER, int(4)),
    (OrsoTypes.DOUBLE, float(4)),
    (OrsoTypes.STRUCT, {"a": "b"}),
    (OrsoTypes.DATE, opteryx.utils.dates.parse_iso("2022-01-01").date()),
    (OrsoTypes.TIMESTAMP, opteryx.utils.dates.parse_iso("2022-01-01 13:31")),
    (OrsoTypes.VARCHAR, "1" * 50),
]


def measure_performance(data_type, op):
    import time

    table = pyarrow.Table.from_arrays([[1] * int(1e6)], ["one"])

    operand = None
    for literal, value in LITERALS:
        if literal == data_type:
            operand = Node(node_type=NodeType.LITERAL, type=data_type, value=value)
            operand = evaluate(operand, table)
            break

    t = time.monotonic_ns()
    try:
        filter_operations(operand, op, operand)
    except:
        return None
    return ((time.monotonic_ns() - t) / 1e9, data_type, op)


def hash_measure_performance(data_type, op):
    import time

    table = pyarrow.Table.from_arrays([[1] * int(1e6)], ["one"])

    operand = None
    for literal, value in LITERALS:
        if literal == data_type:
            operand = Node(node_type=NodeType.LITERAL, type=OrsoTypes.INTEGER, value=hash(value))
            break

    if not operand:
        return None

    operator = Node(node_type=NodeType.COMPARISON_OPERATOR, value=op)
    operator.left = operand
    operator.right = operand

    t = time.monotonic_ns()
    try:
        evaluate(operator, table)
    except:
        return None
    return ((time.monotonic_ns() - t) / 1e9, data_type, op)


for t in OrsoTypes:
    for o in OPERATORS:
        print(measure_performance(t, o))

#        if t == OrsoTypes.VARCHAR:
#            print(hash_measure_performance(t, o))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This is a virtual dataset which is calculated at access time.
"""

from orso.schema import FlatColumn
from orso.schema import RelationSchema
from orso.types import OrsoTypes

__all__ = ("read", "schema")


def read(end_date=None, variables={}):
    import pyarrow

    from opteryx.shared.buffer_pool import BufferPool

    bufferpool = BufferPool()

    lru_hits, lru_misses, lru_evictions, lru_inserts = bufferpool.stats

    pool = bufferpool._memory_pool

    buffer = [
        {"key": "bufferpool_commits", "value": str(pool.commits)},
        {"key": "bufferpool_failed_commits", "value": str(pool.failed_commits)},
        {"key": "bufferpool_reads", "value": str(pool.reads)},
        {"key": "bufferpool_read_locks", "value": str(pool.read_locks)},
        {"key": "bufferpool_compaction_l1", "value": str(pool.l1_compaction)},
        {"key": "bufferpool_compaction_l2", "value": str(pool.l2_compaction)},
        {"key": "bufferpool_releases", "value": str(pool.releases)},
        {"key": "bufferpool_capacity", "value": str(pool.size)},
        {"key": "bufferpool_free", "value": str(pool.available_space())},
        {"key": "bufferpool_items", "value": str(len(pool.used_segments))},
        {"key": "lru_hits", "value": str(lru_hits)},
        {"key": "lru_misses", "value": str(lru_misses)},
        {"key": "lru_evictions", "value": str(lru_evictions)},
        {"key": "lru_inserts", "value": str(lru_inserts)},
    ]

    return pyarrow.Table.from_pylist(buffer)


def schema():
    # fmt:off
    return  RelationSchema(
        name="$statistics",
        columns=[
            FlatColumn(name="key", type=OrsoTypes.VARCHAR),
            FlatColumn(name="value", type=OrsoTypes.VARCHAR),
        ],
    )
    # fmt:on



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This is both a marker and a wrapper for key functionality to support predicate/filter
pushdowns. This is where we a sending filters to the thing that is acquiring the data
for the query. For example sending filters to remote database servers, or to pyarrow
readers. This allows for data to be prefiltered before reaching Opteryx - this is
almost always going to be faster than reading, loading and filtering.

Note that for some file types, although we accept the pushdown, we fake it by reading,
loading and filtering. We do this because we have a single file interface and some
accept filters and others don't so we 'fake' the read-time filtering.
"""

import datetime
from typing import Dict

from orso.types import OrsoTypes

from opteryx.exceptions import NotSupportedError
from opteryx.managers.expression import NodeType
from opteryx.models import Node


class PredicatePushable:
    PUSHABLE_OPS: Dict[str, bool] = {
        "Eq": False,
        "NotEq": False,
        "Gt": False,
        "GtEq": False,
        "Lt": False,
        "LtEq": False,
        "Like": False,
        "NotLike": False,
    }

    OPS_XLAT: Dict[str, str] = {
        "Eq": "=",
        "NotEq": "!=",
        "Gt": ">",
        "GtEq": ">=",
        "Lt": "<",
        "LtEq": "<=",
        "Like": "LIKE",
        "NotLike": "NOT LIKE",
    }

    PUSHABLE_TYPES: set = {t for t in OrsoTypes}

    def can_push(self, operator: Node, types: set = None) -> bool:
        if types and not types.issubset(self.PUSHABLE_TYPES):
            return False
        return self.PUSHABLE_OPS.get(operator.condition.value, False)

    def __init__(self, **kwargs):
        pass

    @staticmethod
    def to_dnf(root):
        """
        Convert a filter to DNF form, this is the form used by PyArrow.

        This is specifically opinionated for the Parquet reader for PyArrow.
        """

        def _predicate_to_dnf(root):
            # Reduce look-ahead effort by using Exceptions to control flow
            if root.node_type == NodeType.AND:  # pragma: no cover
                left = _predicate_to_dnf(root.left)
                right = _predicate_to_dnf(root.right)
                if not isinstance(left, list):
                    left = [left]
                if not isinstance(right, list):
                    right = [right]
                left.extend(right)
                return left
            if root.node_type != NodeType.COMPARISON_OPERATOR:
                raise NotSupportedError()
            if root.left.node_type != NodeType.IDENTIFIER:
                root.left, root.right = root.right, root.left
            if root.right.type == OrsoTypes.DATE:
                date_val = root.right.value
                if hasattr(date_val, "item"):
                    date_val = date_val.item()
                root.right.value = datetime.datetime.combine(date_val, datetime.time.min)
            if root.left.node_type != NodeType.IDENTIFIER:
                raise NotSupportedError()
            if root.right.node_type != NodeType.LITERAL:
                raise NotSupportedError()
            return (
                root.left.value,
                PredicatePushable.OPS_XLAT[root.value],
                root.right.value,
            )

        not_converted = []
        dnf = []
        if not isinstance(root, list):
            root = [root]
        for predicate in root:
            try:
                converted = _predicate_to_dnf(predicate)
                dnf.append(converted)
            except NotSupportedError:
                not_converted.append(predicate)
        return dnf if dnf else None, not_converted



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import typing
from os import environ
from pathlib import Path
from typing import Optional

import psutil

_config_values: dict = {}

# we need a preliminary version of this variable
_OPTERYX_DEBUG = environ.get("OPTERYX_DEBUG") is not None


def memory_allocation_calculation(allocation) -> int:
    """
    Configure the memory allocation for the database based on the input.
    If the allocation is between 0 and 1, it's treated as a percentage of the total system memory.
    If the allocation is greater than 1, it's treated as an absolute value in megabytes.

    Parameters:
        allocation (float|int): Memory allocation value which could be a percentage or an absolute value.

    Returns:
        int: Memory size in bytes to be allocated.
    """
    total_memory = psutil.virtual_memory().total  # Convert bytes to megabytes

    if 0 < allocation < 1:  # Treat as a percentage
        return int(total_memory * allocation)
    elif allocation >= 1:  # Treat as an absolute value in MB
        return int(allocation)
    else:
        raise ValueError("Invalid memory allocation value. Must be a positive number.")


def parse_yaml(yaml_str):
    ## Based on an algorithm from ChatGPT

    def line_value(value):
        value = value.strip()
        if value.isdigit():
            value = int(value)
        elif value.replace(".", "", 1).isdigit():
            value = float(value)
        elif value.lower() == "true":
            return True
        elif value.lower() == "false":
            return False
        elif value.lower() == "none":
            return None
        elif value.startswith("["):
            return [val.strip() for val in value[1:-1].split(",")]
        elif value.startswith("-"):
            return [val.strip() for val in value.split("-") if val.strip()]
        return value

    result: dict = {}
    lines = yaml_str.strip().split("\n")
    key = ""
    value: typing.Any = ""
    in_list = False
    list_key = ""
    for line in lines:
        ## remove comments
        line = line.split("#")[0]
        line = line.strip()
        if not line:
            continue
        if in_list:
            if line.startswith("- "):
                result[list_key].append(line[2:].strip())
            elif line.count(":") == 1:
                key, value = line.split(":", 1)
                if not value.strip():
                    in_list = False
                else:
                    if not isinstance(result[list_key], dict):
                        result[list_key] = {}
                    result[list_key][key.strip()] = line_value(value.strip())
            else:
                if isinstance(result[list_key][0], tuple):
                    result[list_key] = dict(result[list_key])
                in_list = False
        if not in_list:
            key, value = line.split(":", 1)
            if not value.split():
                in_list = True
                list_key = key.strip()
                result[list_key] = []
            else:
                result[key.strip()] = line_value(value)
    return result


try:  # pragma: no cover
    _config_path = Path(".") / "opteryx.yaml"
    if _config_path.exists():
        with open(_config_path, "r") as _config_file:
            _config_values = parse_yaml(_config_file.read())
        if _OPTERYX_DEBUG:
            print(f"{datetime.datetime.now()} [LOADER] Loading config from {_config_path}")
except Exception as exception:  # pragma: no cover # it doesn't matter why - just use the defaults
    if _OPTERYX_DEBUG:
        print(
            f"{datetime.datetime.now()} [LOADER] Config file {_config_path} not used - {exception}"
        )


def get(key, default=None):
    value = environ.get(key)
    if value is None:
        value = _config_values.get(key, default)
    return value


# fmt:off

# These are 'protected' properties which cannot be overridden by a single query

DISABLE_OPTIMIZER: bool = bool(get("DISABLE_OPTIMIZER", False))
"""**DANGEROUS** This will cause most queries to fail."""

OPTERYX_DEBUG: bool = bool(get("OPTERYX_DEBUG", False))
"""Diagnostic and debug mode."""

MAX_CACHE_EVICTIONS_PER_QUERY: int = int(get("MAX_CACHE_EVICTIONS_PER_QUERY", 64))
"""Maximum number of buffer pool evictions by a single query."""

MAX_CACHEABLE_ITEM_SIZE: int = int(get("MAX_CACHEABLE_ITEM_SIZE", 2 * 1024 * 1024))
"""Maximum size for items saved to the remote buffer."""

MAX_LOCAL_BUFFER_CAPACITY: int = memory_allocation_calculation(float(get("MAX_LOCAL_BUFFER_CAPACITY", 0.2)))
"""Local buffer pool size in either bytes or fraction of system memory."""

MAX_READ_BUFFER_CAPACITY: int = memory_allocation_calculation(float(get("MAX_READ_BUFFER_CAPACITY", 0.1)))
"""Read buffer pool size in either bytes or fraction of system memory."""

CONCURRENT_READS: int = int(get("CONCURRENT_READS", 4))
"""Number of read workers per data source."""

DATA_CATALOG_PROVIDER: str = get("DATA_CATALOG_PROVIDER")
"""Data Catalog provider."""

DATA_CATALOG_CONFIGURATION: Optional[str] = get("DATA_CATALOG_CONFIGURATION")
"""Data Catalog configuration, different catalogs have different config formats."""

# GCP project ID - for Google Cloud Data
GCP_PROJECT_ID: str = get("GCP_PROJECT_ID") 
# don't try to raise the priority of the server process
DISABLE_HIGH_PRIORITY: bool = bool(get("DISABLE_HIGH_PRIORITY", False))
# don't output resource (memory) utilization information
ENABLE_RESOURCE_LOGGING: bool = bool(get("ENABLE_RESOURCE_LOGGING", False))
# size of morsels to push between steps
MORSEL_SIZE: int = int(get("MORSEL_SIZE", 64 * 1024 * 1024))
# query log
QUERY_LOG_LOCATION:str = get("QUERY_LOG_LOCATION", False)
QUERY_LOG_SIZE:int = int(get("QUERY_LOG_SIZE", 100))
# not GA
PROFILE_LOCATION:str = get("PROFILE_LOCATION")
# fmt:on



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The Execution Tree is the Graph which defines a Query Plan.

The execution tree contains functionality to:

- build and define the plan
- execute the plan
- manipulate the plan

"""

from typing import Any
from typing import Generator
from typing import Tuple
from typing import Union

import pyarrow

from opteryx.constants import ResultType
from opteryx.exceptions import InvalidInternalStateError
from opteryx.third_party.travers import Graph


class ExecutionTree(Graph):
    """
    The execution tree is defined separately to the planner to simplify the
    complex code which is the planner from the tree that describes the plan.
    """

    def execute(
        self,
    ) -> Generator[Tuple[Union[pyarrow.Table, Any], ResultType], None, None]:
        """
        Implements a 'pull' model execution engine, pulling records starting from
        the last stage (head) of the query plan, and working backwards towards the first stage.

        Yields:
            tuple: The first element is the result (either tabular data or a
                NonTabularResult object). The second element is a ResultType enum,
                indicating the type of the result.
        """
        from opteryx.models import NonTabularResult
        from opteryx.operators import ExplainNode

        def map_operators_to_producers(nodes: list) -> None:
            """
            Walks through the query plan, linking each operator node with its data producers.

            Parameters:
                nodes: list
                    List of operator nodes in the query plan.
            """

            for node in nodes:
                producers = self.ingoing_edges(node)
                operator = self[node]

                if len(producers) == 1:
                    # If there is only one producer, set it directly
                    operator.set_producers([self[producers[0][0]]])
                elif len(producers) == 2 and hasattr(operator, "_left_relation"):
                    left_producer = None
                    right_producer = None

                    left_relation = operator._left_relation
                    right_relation = operator._right_relation
                    for source, target, relation in producers:
                        for s, t, r in self.breadth_first_search(source, reverse=True) + [
                            (source, target, relation)
                        ]:
                            if set(left_relation).intersection(
                                {
                                    self[s].parameters.get("alias"),
                                    self[s].parameters.get("relation"),
                                }
                            ):
                                left_producer = self[source]
                            elif set(right_relation).intersection(
                                {
                                    self[s].parameters.get("alias"),
                                    self[s].parameters.get("relation"),
                                }
                            ):
                                right_producer = self[source]

                    if left_producer and right_producer:
                        operator.set_producers([left_producer, right_producer])
                    else:
                        # Default to setting producers as in the current method if left and right cannot be determined
                        operator.set_producers([self[src_node[0]] for src_node in producers])
                else:
                    # Handle cases with more than two producers if applicable
                    operator.set_producers([self[src_node[0]] for src_node in producers])

                # Recursively process the producers
                map_operators_to_producers([src_node[0] for src_node in producers])

        # Validate query plan to ensure it's acyclic
        if not self.is_acyclic():
            raise InvalidInternalStateError("Query plan is cyclic, cannot execute.")

        # Retrieve the tail of the query plan, which should ideally be a single head node
        head_nodes = list(set(self.get_exit_points()))

        if len(head_nodes) != 1:
            raise InvalidInternalStateError(
                f"Query plan has {len(head_nodes)} heads, expected exactly 1."
            )

        head_node = head_nodes[0]

        # Special case handling for 'Explain' queries
        if isinstance(self[head_node], ExplainNode):
            yield self.explain(), ResultType.TABULAR
            return

        # Link operators with their producers
        map_operators_to_producers([head_node])

        # Execute the head node's operation
        operator = self[head_node]
        results = operator.execute()

        # If the results are non-tabular, handle them accordingly
        if isinstance(results, NonTabularResult):
            yield results, ResultType.NON_TABULAR
        else:
            yield results, ResultType.TABULAR

    def explain(self):
        from opteryx import operators

        def _inner_explain(node, depth):
            incoming_operators = self.ingoing_edges(node)
            for operator_name in incoming_operators:
                operator = self[operator_name[0]]
                if isinstance(
                    operator, (operators.ExitNode, operators.ExplainNode)
                ):  # Skip ExitNode
                    yield from _inner_explain(operator_name[0], depth)
                    continue
                elif isinstance(operator, operators.BasePlanNode):
                    yield {
                        "operator": operator.name,
                        "config": operator.config,
                        "depth": depth,
                    }
                    yield from _inner_explain(operator_name[0], depth + 1)

        head = list(dict.fromkeys(self.get_exit_points()))
        if len(head) != 1:  # pragma: no cover
            raise InvalidInternalStateError(f"Problem with the plan - it has {len(head)} heads.")
        plan = list(_inner_explain(head[0], 1))

        table = pyarrow.Table.from_pylist(plan)

        yield table
